"""
Native storage for OmegaML using mongodb as the storage layer

An OmegaStore instance is a MongoDB database. It has at least the
metadata collection which lists all objects stored in it. A metadata
document refers to the following types of objects (metadata.kind):

* pandas.dfrows - a Pandas DataFrame stored as a collection of rows
* sklearn.joblib - a scikit learn estimator/pipline dumped using joblib.dump()
* python.data - an arbitrary python dict, tuple, list stored as a document

Note that storing Pandas and scikit learn objects requires the availability
of the respective packages. If either can not be imported, the OmegaStore
degrades to a python.data store only. It will still .list() and get() any
object, however reverts to pure python objects. In this case it is up
to the client to convert the data into an appropriate format for processing.

Pandas and scikit-learn objects can only be stored if these packages are
availables. put() raises a TypeError if you pass such objects and these
modules cannot be loaded.

All data are stored within the same mongodb, in per-object collections
as follows:

    * .metadata
        all metadata. each object is one document,
        See **omegaml.documents.Metadata** for details
    * .<bucket>.files
        this is the GridFS instance used to store
        blobs (models, numpy, hdf). The actual file name
        will be <prefix>/<name>.<ext>, where ext is
        optionally generated by put() / get().
    * .<bucket>.<prefix>.<name>.data
        every other dataset is stored in a separate
        collection (dataframes, dicts, lists, tuples).
        Any forward slash in prefix is ignored (e.g. 'data/'
        becomes 'data')

    DataFrames by default are stored in their own collection, every
    row becomes a document. To store dataframes as a binary file,
    use `put(...., as_hdf=True).` `.get()` will always return a dataframe.

    Python dicts, lists, tuples are stored as a single document with
    a `.data` attribute holding the JSON-converted representation. `.get()`
    will always return the corresponding python object of .data.

    Models are joblib.dump()'ed and ziped prior to transferring into
    GridFs. .get() will always unzip and joblib.load() before returning
    the model. Note this requires that the process using .get() supports
    joblib as well as all python classes referred to. If joblib is not
    supported, .get() returns a file-like object.

    The .metadata entry specifies the format used to store each
    object as well as it's location:

    * metadata.kind
        the type of object
    * metadata.name
        the name of the object, as given on put()
    * metadata.gridfile
        the gridfs object (if any, null otherwise)
    * metadata.collection
        the name of the collection
    * metadata.attributes
        arbitrary custom attributes set in
        put(attributes=obj). This is used e.g. by
        OmegaRuntime's fit() method to record the
        data used in the model's training.

    **.put()** and **.get()** use helper methods specific to the type in
    object's type and metadata.kind, respectively. In the future
    a plugin system will enable extension to other types.
"""
from __future__ import absolute_import

import os
import re
import tempfile
import warnings
from fnmatch import fnmatch
from uuid import uuid4

import gridfs
import six
from datetime import datetime
from mongoengine.connection import disconnect, \
    connect
from mongoengine.errors import DoesNotExist
from mongoengine.fields import GridFSProxy
from six import iteritems

from omegaml.store.fastinsert import fast_insert
from omegaml.util import unravel_index, restore_index, make_tuple, jsonescape, \
    cursor_to_dataframe, convert_dtypes, load_class, extend_instance
from ..documents import make_Metadata, MDREGISTRY
from ..util import (is_estimator, is_dataframe, is_ndarray, is_spark_mllib,
                    settings as omega_settings, urlparse, is_series)


class OmegaStore(object):
    """
    The storage backend for models and data
    """

    def __init__(self, mongo_url=None, bucket=None, prefix=None, kind=None, defaults=None):
        """
        :param mongo_url: the mongourl to use for the gridfs
        :param bucket: the mongo collection to use for gridfs
        :param prefix: the path prefix for files. defaults to blank
        :param kind: the kind or list of kinds to limit this store to
        """
        self.defaults = defaults or omega_settings()
        self.mongo_url = mongo_url or self.defaults.OMEGA_MONGO_URL
        self.bucket = bucket or self.defaults.OMEGA_MONGO_COLLECTION
        self._fs = None
        self._tmppath = None
        self.prefix = prefix or ''
        self.force_kind = kind
        self._Metadata_cls = None
        # don't initialize db here to avoid using the default settings
        # otherwise Metadata will already have a connection and not use
        # the one provided in override_settings
        self._db = None
        # add backends and mixins
        self._apply_mixins()
        # register backends
        self.register_backends()

    def __repr__(self):
        return 'OmegaStore(bucket={}, prefix={})'.format(self.bucket, self.prefix)

    def __equal__(self, other):
        """test for equality of OmegaStore instances

        Args:
            other: OmegaStore instance

        Returns:
            True if other is the same database, same bucket, same prefix
        """
        return self.mongo_url == other.mongo_url and self.bucket == other.bucket and self.prefix == other.prefix

    @property
    def tmppath(self):
        """
        return an instance-specific temporary path
        """
        if self._tmppath is not None:
            return self._tmppath
        self._tmppath = os.path.join(self.defaults.OMEGA_TMP, uuid4().hex)
        os.makedirs(self._tmppath)
        return self._tmppath

    @property
    def mongodb(self):
        """
        Returns a mongo database object
        """
        if self._db is not None:
            return self._db
        # parse salient parts of mongourl, e.g.
        # mongodb://user:password@host/dbname
        self.parsed_url = urlparse.urlparse(self.mongo_url)
        self.database_name = self.parsed_url.path[1:]
        host = self.parsed_url.netloc
        username, password = None, None
        if '@' in host:
            creds, host = host.split('@', 1)
            if ':' in creds:
                username, password = creds.split(':')
        # connect via mongoengine
        # note this uses a MongoClient in the background, with pooled
        # connections. there are multiprocessing issues with pymongo:
        # http://api.mongodb.org/python/3.2/faq.html#using-pymongo-with-multiprocessing
        # connect=False is due to https://jira.mongodb.org/browse/PYTHON-961
        # this defers connecting until the first access
        # serverSelectionTimeoutMS=2500 is to fail fast, the default is 30000
        # FIXME use an instance specific alias. requires that every access
        #       to Metadata is configured correctly. this to avoid sharing
        #       inadevertedly between threads and processes.
        # alias = 'omega-{}'.format(uuid4().hex)
        alias = 'omega'
        # always disconnect before registering a new connection because
        # connect forgets all connection settings upon disconnect WTF?!
        disconnect(alias)
        connection = connect(alias=alias, db=self.database_name,
                             host=host,
                             username=username,
                             password=password,
                             connect=False,
                             authentication_source='admin',
                             serverSelectionTimeoutMS=2500,
                             **self.defaults.OMEGA_MONGO_SSL_KWARGS,
                            )
        self._db = getattr(connection, self.database_name)
        # mongoengine 0.15.0 connection setup is seriously broken -- it does
        # not remember username/password on authenticated connections
        # so we reauthenticate here
        if username and password:
            self._db.logout()
            self._db.authenticate(username, password, source='admin')
        return self._db

    @property
    def _Metadata(self):
        if self._Metadata_cls is None:
            # hack to localize metadata
            self.mongodb
            self._Metadata_cls = make_Metadata()
        return self._Metadata_cls

    @property
    def fs(self):
        """
        Retrieve a gridfs instance using url and collection provided

        :return: a gridfs instance
        """
        if self._fs is not None:
            return self._fs
        try:
            self._fs = gridfs.GridFS(self.mongodb, collection=self.bucket)
        except Exception as e:
            raise e
        return self._fs

    def metadata(self, name=None, bucket=None, prefix=None, version=-1):
        """
        Returns a metadata document for the given entry name

        FIXME: version attribute does not do anything
        FIXME: metadata should be stored in a bucket-specific collection
        to enable access control, see https://docs.mongodb.com/manual/reference/method/db.createRole/#db.createRole
        """
        db = self.mongodb
        fs = self.fs
        prefix = prefix or self.prefix
        bucket = bucket or self.bucket
        # Meta is to silence lint on import error
        Meta = self._Metadata
        return Meta.objects(name=name, prefix=prefix, bucket=bucket).no_cache().first()

    def make_metadata(self, name, kind, bucket=None, prefix=None, **kwargs):
        """
        create or update a metadata object

        this retrieves a Metadata object if it exists given the kwargs. Only
        the name, prefix and bucket arguments are considered

        for existing Metadata objects, the attributes kw is treated as follows:

        * attributes=None, the existing attributes are left as is
        * attributes={}, the attributes value on an existing metadata object
          is reset to the empty dict
        * attributes={ some : value }, the existing attributes are updated

        For new metadata objects, attributes defaults to {} if not specified,
        else is set as provided.

        :param name: the object name
        :param bucket: the bucket, optional, defaults to self.bucket
        :param prefix: the prefix, optional, defaults to self.prefix

        """
        # TODO kept _make_metadata for backwards compatibility.
        return self._make_metadata(name, bucket=bucket, prefix=prefix,
                                   kind=kind, **kwargs)

    def _make_metadata(self, name=None, bucket=None, prefix=None, **kwargs):
        """
        create or update a metadata object

        this retrieves a Metadata object if it exists given the kwargs. Only
        the name, prefix and bucket arguments are considered

        for existing Metadata objects, the attributes kw is treated as follows:

        * attributes=None, the existing attributes are left as is
        * attributes={}, the attributes value on an existing metadata object
        is reset to the empty dict
        * attributes={ some : value }, the existing attributes are updated

        For new metadata objects, attributes defaults to {} if not specified,
        else is set as provided.

        :param name: the object name
        :param bucket: the bucket, optional, defaults to self.bucket
        :param prefix: the prefix, optional, defaults to self.prefix
        """
        bucket = bucket or self.bucket
        prefix = prefix or self.prefix
        meta = self.metadata(name=name,
                             prefix=prefix,
                             bucket=bucket)
        if meta:
            dict_fields = 'attributes', 'kind_meta'
            for k, v in six.iteritems(kwargs):
                if k in dict_fields and v is not None and len(v) > 0:
                    previous = getattr(meta, k, {})
                    previous.update(v)
                    setattr(meta, k, previous)
                elif k in dict_fields and v is not None and len(v) == 0:
                    setattr(meta, k, {})
                elif k in dict_fields and v is None:
                    # ignore non specified attributes
                    continue
                else:
                    # by default set whatever attribute is provided
                    setattr(meta, k, v)
        else:
            meta = self._Metadata(name=name, bucket=bucket, prefix=prefix,
                                  **kwargs)
        return meta

    def _drop_metadata(self, name=None, **kwargs):
        # internal method to delete meta data of an object
        meta = self.metadata(name, **kwargs)
        if meta is not None:
            meta.delete()

    def collection(self, name=None, bucket=None, prefix=None):
        """
        Returns a mongo db collection as a datastore

        If there is an existing object of name, will return the .collection
        of the object. Otherwise returns the collection according to naming
        convention {bucket}.{prefix}.{name}.datastore

        :param name: the collection to use. if none defaults to the
            collection name given on instantiation. the actual collection name
            used is always prefix + name + '.data'
        """
        # see if we have a known object and a collection for that, if not define it
        meta = self.metadata(name, bucket=bucket, prefix=prefix)
        collection = meta.collection if meta else None
        if not collection:
            collection = self._get_obj_store_key(name, '.datastore')
            collection = collection.replace('..', '.')
        # return the collection
        try:
            datastore = getattr(self.mongodb, collection)
        except Exception as e:
            raise e
        return datastore

    def _apply_mixins(self):
        """
        apply mixins in defaults.OMEGA_STORE_MIXINS
        """
        for mixin in self.defaults.OMEGA_STORE_MIXINS:
            extend_instance(self, mixin)

    def register_backends(self):
        """
        register backends in defaults.OMEGA_STORE_BACKENDS
        """
        for kind, backend in six.iteritems(self.defaults.OMEGA_STORE_BACKENDS):
            self.register_backend(kind, backend)

    def register_backend(self, kind, backend):
        """
        register a backend class

        :param kind: (str) the backend kind
        :param backend: (class) the backend class
        """
        self.defaults.OMEGA_STORE_BACKENDS[kind] = load_class(backend)
        if kind not in MDREGISTRY.KINDS:
            MDREGISTRY.KINDS.append(kind)
        return self

    def register_mixin(self, mixincls):
        """
        register a mixin class

        :param mixincls: (class) the mixin class
        """
        self.defaults.OMEGA_STORE_MIXINS.append(mixincls)
        extend_instance(self, mixincls)
        return self

    def put(self, obj, name, attributes=None, kind=None, **kwargs):
        """
        Stores an object, store estimators, pipelines, numpy arrays or
        pandas dataframes
        """
        backend = self.get_backend_byobj(obj, name, attributes=attributes, kind=kind, **kwargs)
        if backend:
            return backend.put(obj, name, attributes=attributes, **kwargs)
        # TODO move all of the specifics to backend implementations
        if is_estimator(obj):
            backend = self.get_backend_bykind(MDREGISTRY.SKLEARN_JOBLIB)
            return backend.put_model(obj, name, attributes)
        elif is_spark_mllib(obj):
            backend = self.get_backend_bykind(MDREGISTRY.SKLEARN_JOBLIB)
            return backend.put_model(obj, name, attributes, **kwargs)
        elif is_dataframe(obj) or is_series(obj):
            groupby = kwargs.get('groupby')
            if obj.empty:
                from warnings import warn
                warn(
                    'Provided dataframe is empty, ignoring it, doing nothing here!')
                return None
            if kwargs.pop('as_hdf', False):
                return self.put_dataframe_as_hdf(
                    obj, name, attributes, **kwargs)
            elif groupby:
                return self.put_dataframe_as_dfgroup(
                    obj, name, groupby, attributes)
            append = kwargs.get('append', None)
            timestamp = kwargs.get('timestamp', None)
            index = kwargs.get('index', None)
            return self.put_dataframe_as_documents(
                obj, name, append, attributes, index, timestamp)
        elif is_ndarray(obj):
            return self.put_ndarray_as_hdf(obj, name,
                                           attributes=attributes,
                                           **kwargs)
        elif isinstance(obj, (dict, list, tuple)):
            if kwargs.pop('as_hdf', False):
                self.put_pyobj_as_hdf(obj, name,
                                      attributes=attributes, **kwargs)
            return self.put_pyobj_as_document(obj, name,
                                              attributes=attributes,
                                              **kwargs)
        else:
            raise TypeError('type %s not supported' % type(obj))

    def put_dataframe_as_documents(self, obj, name, append=None,
                                   attributes=None, index=None,
                                   timestamp=None):
        """
        store a dataframe as a row-wise collection of documents

        :param obj: the dataframe to store
        :param name: the name of the item in the store
        :param append: if False collection will be dropped before inserting,
           if True existing documents will persist. Defaults to True. If not
           specified and rows have been previously inserted, will issue a
           warning.
        :param index: list of columns, using +, -, @ as a column prefix to
           specify ASCENDING, DESCENDING, GEOSPHERE respectively. For @ the
           column has to represent a valid GeoJSON object.
        :param timestamp: if True or a field name adds a timestamp. If the
           value is a boolean or datetime, uses _created as the field name.
           The timestamp is always datetime.datetime.utcnow(). May be overriden
           by specifying the tuple (col, datetime).
        :return: the Metadata object created
        """
        from .queryops import MongoQueryOps
        collection = self.collection(name)
        if is_series(obj):
            import pandas as pd
            obj = pd.DataFrame(obj, index=obj.index, columns=[str(obj.name)])
            store_series = True
        else:
            store_series = False
        if append is False:
            self.drop(name, force=True)
        elif append is None and collection.count(limit=1):
            from warnings import warn
            warn('%s already exists, will append rows' % name)
        if index:
            # get index keys
            if isinstance(index, dict):
                idx_kwargs = index
                index = index.pop('columns')
            else:
                idx_kwargs = {}
            # create index with appropriate options
            keys, idx_kwargs = MongoQueryOps().make_index(index, **idx_kwargs)
            collection.create_index(keys, **idx_kwargs)
        if timestamp:
            dt = datetime.utcnow()
            if isinstance(timestamp, bool):
                col = '_created'
            elif isinstance(timestamp, six.string_types):
                col = timestamp
            elif isinstance(timestamp, datetime):
                col, dt = '_created', timestamp
            elif isinstance(timestamp, tuple):
                col, dt = timestamp
            obj[col] = dt
        # store dataframe indicies
        # FIXME this may be a performance issue, use size stored on stats or metadata
        row_count = self.collection(name).count()
        obj, idx_meta = unravel_index(obj, row_count=row_count)
        stored_columns = [jsonescape(col) for col in obj.columns]
        column_map = list(zip(obj.columns, stored_columns))
        dtypes = {
            dict(column_map).get(k): v.name
            for k, v in iteritems(obj.dtypes)
            }
        kind_meta = {
            'columns': column_map,
            'dtypes': dtypes,
            'idx_meta': idx_meta
        }
        # ensure column names to be strings
        obj.columns = stored_columns
        # create mongon indicies for data frame index columns
        df_idxcols = [col for col in obj.columns if col.startswith('_idx')]
        if df_idxcols:
            keys, idx_kwargs = MongoQueryOps().make_index(df_idxcols)
            collection.create_index(keys, **idx_kwargs)
        # create index on row id
        keys, idx_kwargs = MongoQueryOps().make_index(['_om#rowid'])
        collection.create_index(keys, **idx_kwargs)
        # bulk insert
        # -- get native objects
        # -- seems to be required since pymongo 3.3.x. if not converted
        #    pymongo raises Cannot Encode object for int64 types
        obj = obj.astype('O')
        fast_insert(obj, self, name)
        kind = (MDREGISTRY.PANDAS_SEROWS
                if store_series
                else MDREGISTRY.PANDAS_DFROWS)
        meta = self._make_metadata(name=name,
                                   prefix=self.prefix,
                                   bucket=self.bucket,
                                   kind=kind,
                                   kind_meta=kind_meta,
                                   attributes=attributes,
                                   collection=collection.name)
        return meta.save()

    def put_dataframe_as_dfgroup(self, obj, name, groupby, attributes=None):
        """
        store a dataframe grouped by columns in a mongo document

        :Example:

          > # each group
          >  {
          >     #group keys
          >     key: val,
          >     _data: [
          >      # only data keys
          >        { key: val, ... }
          >     ]}

        """

        def row_to_doc(obj):
            for gval, gdf in obj.groupby(groupby):
                if hasattr(gval,'astype'):
                    gval = make_tuple(gval.astype('O'))
                else:
                    gval = make_tuple(gval)
                doc = dict(zip(groupby, gval))
                datacols = list(set(gdf.columns) - set(groupby))
                doc['_data'] = gdf[datacols].astype('O').to_dict('records')
                yield doc

        datastore = self.collection(name)
        datastore.drop()
        datastore.insert_many(row_to_doc(obj))
        return self._make_metadata(name=name,
                                   prefix=self.prefix,
                                   bucket=self.bucket,
                                   kind=MDREGISTRY.PANDAS_DFGROUP,
                                   attributes=attributes,
                                   collection=datastore.name).save()

    def put_dataframe_as_hdf(self, obj, name, attributes=None):
        filename = self._get_obj_store_key(name, '.hdf')
        hdffname = self._package_dataframe2hdf(obj, filename)
        with open(hdffname, 'rb') as fhdf:
            fileid = self.fs.put(fhdf, filename=filename)
        return self._make_metadata(name=name,
                                   prefix=self.prefix,
                                   bucket=self.bucket,
                                   kind=MDREGISTRY.PANDAS_HDF,
                                   attributes=attributes,
                                   gridfile=GridFSProxy(grid_id=fileid)).save()

    def put_ndarray_as_hdf(self, obj, name, attributes=None):
        """ store numpy array as hdf

        this is hack, converting the array to a dataframe then storing
        it
        """
        import pandas as pd
        df = pd.DataFrame(obj)
        return self.put_dataframe_as_hdf(df, name, attributes=attributes)

    def put_pyobj_as_hdf(self, obj, name, attributes=None):
        """
        store list, tuple, dict as hdf

        this requires the list, tuple or dict to be convertible into
        a dataframe
        """
        import pandas as pd
        df = pd.DataFrame(obj)
        return self.put_dataframe_as_hdf(df, name, attributes=attributes)

    def put_pyobj_as_document(self, obj, name, attributes=None, append=True):
        """
        store a dict as a document

        similar to put_dataframe_as_documents no data will be replaced by
        default. that is, obj is appended as new documents into the objects'
        mongo collection. to replace the data, specify append=False.
        """
        collection = self.collection(name)
        if append is False:
            collection.drop()
        elif append is None and collection.count(limit=1):
            from warnings import warn
            warn('%s already exists, will append rows' % name)
        objid = collection.insert({'data': obj})
        return self._make_metadata(name=name,
                                   prefix=self.prefix,
                                   bucket=self.bucket,
                                   kind=MDREGISTRY.PYTHON_DATA,
                                   collection=collection.name,
                                   attributes=attributes,
                                   objid=objid).save()

    def drop(self, name, force=False, version=-1):
        """
        Drop the object

        :param name: The name of the object
        :param force: If True ignores DoesNotExist exception, defaults to False
            meaning this raises a DoesNotExist exception of the name does not
            exist
        :return:    True if object was deleted, False if not.
                    If force is True and
                    the object does not exist it will still return True
        """
        meta = self.metadata(name, version=version)
        if meta is None and not force:
            raise DoesNotExist()
        collection = self.collection(name)
        if collection:
            self.mongodb.drop_collection(collection.name)
        if meta:
            if meta.collection:
                self.mongodb.drop_collection(meta.collection)
                self._drop_metadata(name)
                return True
            if meta and meta.gridfile is not None:
                meta.gridfile.delete()
                self._drop_metadata(name)
                return True
        return False

    def get_backend_bykind(self, kind, model_store=None, data_store=None,
                           **kwargs):
        """
        return the backend by a given object kind

        :param kind: The object kind
        :param model_store: the OmegaStore instance used to store models
        :param data_store: the OmegaStore instance used to store data
        :param kwargs: the kwargs passed to the backend initialization
        :return: the backend
        """
        try:
            backend_cls = load_class(self.defaults.OMEGA_STORE_BACKENDS[kind])
        except KeyError as e:
            raise ValueError('backend {kind} does not exist'.format(**locals()))
        model_store = model_store or self
        data_store = data_store or self
        backend = backend_cls(model_store=model_store,
                              data_store=data_store, **kwargs)
        return backend

    def get_backend(self, name, model_store=None, data_store=None, **kwargs):
        """
        return the backend by a given object name

        :param kind: The object kind
        :param model_store: the OmegaStore instance used to store models
        :param data_store: the OmegaStore instance used to store data
        :param kwargs: the kwargs passed to the backend initialization
        :return: the backend
        """
        meta = self.metadata(name)
        if meta is not None and meta.kind in self.defaults.OMEGA_STORE_BACKENDS:
            return self.get_backend_bykind(meta.kind,
                                           model_store=model_store,
                                           data_store=data_store)
        return None

    def get_backend_byobj(self, obj, name, kind=None, attributes=None,
                          model_store=None, data_store=None, **kwargs):
        """
        return the matching backend for the given obj

        Returns:
            the first backend that supports the given parameters or None
        """
        model_store = model_store or self
        data_store = data_store or self
        meta = self.metadata(name)
        kind = kind or (meta.kind if meta is not None else None)
        if kind and kind in self.defaults.OMEGA_STORE_BACKENDS:
            backend = self.get_backend_bykind(kind)
            if not backend.supports(obj, name, attributes=attributes,
                                    data_store=data_store,
                                    model_store=model_store, **kwargs):
                objtype = str(type(obj))
                warnings.warn('Backend {kind} does not support {objtype}'.format(**locals()))
            return backend
        for kind, backend_cls in six.iteritems(self.defaults.OMEGA_STORE_BACKENDS):
            backend = self.get_backend_bykind(kind)
            if backend.supports(obj, name, attributes=attributes, **kwargs):
                return backend
        return None

    def getl(self, *args, **kwargs):
        """ return a lazy MDataFrame for a given object

        Same as .get, but returns a MDataFrame

        """
        return self.get(*args, lazy=True, **kwargs)

    def get(self, name, version=-1, force_python=False,
            kind=None, **kwargs):
        """
        Retrieve an object

        :param name: The name of the object
        :param version: Version of the stored object (not supported)
        :param force_python: Return as a python object
        :param kwargs: kwargs depending on object kind
        :return: an object, estimator, pipelines, data array or pandas dataframe
            previously stored with put()
        """
        meta = self.metadata(name, version=version)
        if meta is None:
            return None
        if not force_python:
            backend = self.get_backend(name) if not kind else self.get_backend_bykind(kind)
            if backend is not None:
                return backend.get(name, **kwargs)
            if meta.kind == MDREGISTRY.SKLEARN_JOBLIB:
                backend = self.get_backend(name)
                return backend.get_model(name)
            elif meta.kind == MDREGISTRY.SPARK_MLLIB:
                backend = self.get_backend(name)
                return backend.get_model(name, version)
            elif meta.kind == MDREGISTRY.PANDAS_DFROWS:
                return self.get_dataframe_documents(name, version=version,
                                                    **kwargs)
            elif meta.kind == MDREGISTRY.PANDAS_SEROWS:
                return self.get_dataframe_documents(name, version=version,
                                                    is_series=True,
                                                    **kwargs)
            elif meta.kind == MDREGISTRY.PANDAS_DFGROUP:
                return self.get_dataframe_dfgroup(
                    name, version=version, **kwargs)
            elif meta.kind == MDREGISTRY.PYTHON_DATA:
                return self.get_python_data(name, version=version, **kwargs)
            elif meta.kind == MDREGISTRY.PANDAS_HDF:
                return self.get_dataframe_hdf(name, version=version)
        return self.get_object_as_python(meta, version=version)

    def get_dataframe_documents(self, name, columns=None, lazy=False,
                                filter=None, version=-1, is_series=False,
                                **kwargs):
        """
        Internal method to return DataFrame from documents

        :param name: the name of the object (str)
        :param columns: the column projection as a list of column names
        :param lazy: if True returns a lazy representation as an MDataFrame.
           If False retrieves all data and returns a DataFrame (default)
        :param filter: the filter to be applied as a column__op=value dict
        :param version: the version to retrieve (not supported)
        :param is_series: if True retruns a Series instead of a DataFrame
        :param kwargs: remaining kwargs are used a filter. The filter kwarg
           overrides other kwargs.
        :return: the retrieved object (DataFrame, Series or MDataFrame)

        """
        collection = self.collection(name)
        if lazy:
            from ..mdataframe import MDataFrame
            filter = filter or kwargs
            df = MDataFrame(collection, columns=columns).query(**filter)
            if is_series:
                df = df[0]
        else:
            # TODO ensure the same processing is applied in MDataFrame
            # TODO this method should always use a MDataFrame disregarding lazy
            filter = filter or kwargs
            if filter:
                from .query import Filter
                query = Filter(collection, **filter).query
                cursor = collection.find(filter=query, projection=columns)
            else:
                cursor = collection.find(projection=columns)
            # restore dataframe
            df = cursor_to_dataframe(cursor)
            if '_id' in df.columns:
                del df['_id']
            meta = self.metadata(name)
            if hasattr(meta, 'kind_meta'):
                df = convert_dtypes(df, meta.kind_meta.get('dtypes', {}))
            # -- restore columns
            meta_columns = dict(meta.kind_meta.get('columns'))
            if meta_columns:
                # apply projection, if any
                if columns:
                    # get only projected columns
                    # meta_columns is {origin_column: stored_column}
                    orig_columns = dict({k: v for k, v in iteritems(meta_columns)
                                         if k in columns or v in columns})
                else:
                    # restore columns to original name
                    orig_columns = meta_columns
                df.rename(columns=orig_columns, inplace=True)
            # -- restore indexes
            idx_meta = meta.kind_meta.get('idx_meta')
            if idx_meta:
                df = restore_index(df, idx_meta)
            # -- restore row order
            if is_series:
                index = df.index
                name = df.columns[0]
                df = df[name]
                df.index = index
                df.name = None if name == 'None' else name
        return df

    def rebuild_params(self, kwargs, collection):
        """
        Returns a modified set of parameters for querying mongodb
        based on how the mongo document is structured and the
        fields the document is grouped by.

        **Note: Explicitly to be used with get_grouped_data only**

        :param kwargs: Mongo filter arguments
        :param collection: The name of mongodb collection
        :return: Returns a set of parameters as dictionary.
        """
        modified_params = {}
        db_structure = collection.find_one({}, {'_id': False})
        groupby_columns = list(set(db_structure.keys()) - set(['_data']))
        if kwargs is not None:
            for item in kwargs:
                if item not in groupby_columns:
                    modified_query_param = '_data.' + item
                    modified_params[modified_query_param] = kwargs.get(item)
                else:
                    modified_params[item] = kwargs.get(item)
        return modified_params

    def get_dataframe_dfgroup(self, name, version=-1, kwargs=None):
        """
        Return a grouped dataframe

        :param name: the name of the object
        :param version: not supported
        :param kwargs: mongo db query arguments to be passed to
               collection.find() as a filter.

        """
        import pandas as pd
        def convert_doc_to_row(cursor):
            for doc in cursor:
                data = doc.pop('_data', [])
                for row in data:
                    doc.update(row)
                    yield doc

        datastore = self.collection(name)
        kwargs = kwargs if kwargs else {}
        params = self.rebuild_params(kwargs, datastore)
        cursor = datastore.find(params, {'_id': False})
        df = pd.DataFrame(convert_doc_to_row(cursor))
        return df

    def get_dataframe_hdf(self, name, version=-1):
        """
        Retrieve dataframe from hdf

        :param name: The name of object
        :param version: The version of object (not supported)
        :return: Returns a python pandas dataframe
        :raises: gridfs.errors.NoFile
        """
        df = None
        filename = self._get_obj_store_key(name, '.hdf')
        if filename.endswith('.hdf') and self.fs.exists(filename=filename):
            df = self._extract_dataframe_hdf(filename, version=version)
            return df
        else:
            raise gridfs.errors.NoFile(
                "{0} does not exist in mongo collection '{1}'".format(
                    name, self.bucket))

    def get_python_data(self, name, version=-1,  **kwargs):
        """
        Retrieve objects as python data

        :param name: The name of object
        :param version: The version of object

        :return: Returns the object as python list object
        """
        datastore = self.collection(name)
        cursor = datastore.find(**kwargs)
        data = (d.get('data') for d in cursor)
        return list(data)

    def get_object_as_python(self, meta, version=-1):
        """
        Retrieve object as python object

        :param meta: The metadata object
        :param version: The version of the object

        :return: Returns data as python object
        """
        if meta.kind == MDREGISTRY.SKLEARN_JOBLIB:
            return meta.gridfile
        if meta.kind == MDREGISTRY.PANDAS_HDF:
            return meta.gridfile
        if meta.kind == MDREGISTRY.PANDAS_DFROWS:
            return list(getattr(self.mongodb, meta.collection).find())
        if meta.kind == MDREGISTRY.PYTHON_DATA:
            col = getattr(self.mongodb, meta.collection)
            return col.find_one(dict(_id=meta.objid)).get('data')
        raise TypeError('cannot return kind %s as a python object' % meta.kind)

    def __iter__(self):
        for f in self.list():
            yield f

    def list(self, pattern=None, regexp=None, kind=None, raw=False, hidden=None,
             include_temp=False, bucket=None, prefix=None, filter=None):
        """
        List all files in store

        specify pattern as a unix pattern (e.g. :code:`models/*`,
        or specify regexp)

        :param pattern: the unix file pattern or None for all
        :param regexp: the regexp. takes precedence over pattern
        :param raw: if True return the meta data objects
        :param filter: specify additional filter criteria, optional
        :return: List of files in store

        """
        db = self.mongodb
        searchkeys = dict(bucket=bucket or self.bucket,
                          prefix=prefix or self.prefix)
        if kind or self.force_kind:
            kind = kind or self.force_kind
            if isinstance(kind, (tuple, list)):
                searchkeys.update(kind__in=kind)
            else:
                searchkeys.update(kind=kind)
        if filter:
            searchkeys.update(filter)
        meta = self._Metadata.objects.no_cache()(**searchkeys)
        if raw:
            if regexp:
                files = [f for f in meta if re.match(regexp, f.name)]
            elif pattern:
                files = [f for f in meta if fnmatch(f.name, pattern)]
            else:
                files = [f for f in meta]
            if not include_temp:
                files = [f for f in files if not f.name.startswith('_')]
            if not hidden:
                files = [f for f in files if not f.name.startswith('.')]
        else:
            files = [str(d.name) for d in meta]
            if regexp:
                files = [f for f in files if re.match(regexp, f)]
            elif pattern:
                files = [f for f in files if fnmatch(f, pattern)]
            files = [f.replace('.omm', '') for f in files]
            if not include_temp:
                files = [f for f in files if not f.startswith('_')]
            if not hidden:
                files = [f for f in files if not f.startswith('.')]
        return files

    def object_store_key(self, name, ext):
        """
        Returns the store key

        :param name: The name of object
        :param ext: The extension of the filename

        :return: A filename with relative bucket, prefix and name
        """
        return self._get_obj_store_key(name, ext)

    def _get_obj_store_key(self, name, ext, prefix=None, bucket=None):
        # backwards compatilibity implementation of object_store_key()
        name = '%s.%s' % (name, ext) if not name.endswith(ext) else name
        filename = '{bucket}.{prefix}.{name}'.format(
            bucket=bucket or self.bucket,
            prefix=prefix or self.prefix,
            name=name,
            ext=ext).replace('/', '_').replace('..', '.')
        return filename

    def _package_dataframe2hdf(self, df, filename, key=None):
        """
        Package a dataframe as a hdf file

        :param df: The dataframe
        :param filename: Name of file

        :return: Filename of hdf file
        """
        lpath = tempfile.mkdtemp()
        fname = os.path.basename(filename)
        hdffname = os.path.join(self.tmppath, fname + '.hdf')
        key = key or 'data'
        df.to_hdf(hdffname, key)
        return hdffname

    def _extract_dataframe_hdf(self, filename, version=-1):
        """
        Extracts a dataframe from a stored hdf file

        :param filename: The name of file
        :param version: The version of file

        :return: Pandas dataframe
        """
        import pandas as pd
        hdffname = os.path.join(self.tmppath, filename)
        dirname = os.path.dirname(hdffname)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        try:
            outf = self.fs.get_version(filename, version=version)
        except gridfs.errors.NoFile as e:
            raise e
        with open(hdffname, 'wb') as hdff:
            hdff.write(outf.read())
        hdf = pd.HDFStore(hdffname)
        key = list(hdf.keys())[0]
        df = hdf[key]
        hdf.close()
        return df



from __future__ import absolute_import
'''
make sure Celery is correctly configured
see http://chriskief.com/2013/11/15/celery-3-1-with-django-django-celery-rabbitmq-and-macports/
'''

import os

from celery import Celery
from omegaml.util import settings

# get rid of celery's Django compatibility mode
os.environ['DJANGO_SETTINGS_MODULE'] = ''

defaults = settings()
app = Celery('omegaml')
app.config_from_object(defaults.OMEGA_CELERY_CONFIG)
app.autodiscover_tasks(defaults.OMEGA_CELERY_IMPORTS, related_name='tasks')



from time import sleep

import requests

from omegaml.client.auth import OmegaRestApiAuth
from omegaml.client.docoptparser import CommandBase
from omegaml.client.userconf import save_userconfig_from_apikey
from omegaml.client.util import get_omega
from omegaml.defaults import update_from_config


class CloudCommandBase(CommandBase):
    """
    Usage:
      om cloud login [<userid>] [<apikey>] [options]
      om cloud config [options]
      om cloud (add|update|remove) <kind> [--node-type <type>] [--count <n>] [options]

    Options:
      --userid=USERID   the userid at hub.omegaml.io (see account profile)
      --apikey=APIKEY   the apikey at hub.omegaml.io (see account profile)
      --apiurl=URL      the cloud URL [default: https://hub.omegaml.io]
      --count=NUMBER    how many instances to set up [default: 1]
      --node-type=TYPE  the type of node [default: small]
    """
    command = 'cloud'

    @property
    def om(self):
        return get_omega(self.args)

    def login(self):
        userid = self.args.get('<userid>') or self.args.get('--userid')
        apikey = self.args.get('<apikey>') or self.args.get('--apikey')
        api_url = self.args.get('--apiurl')
        configfile = self.args.get('--config') or 'config.yml'
        if not userid:
            userid = self.ask('Userid:')
        if not apikey:
            apikey = self.ask('Apikey:')
        save_userconfig_from_apikey(configfile, userid, apikey, api_url=api_url)

    def config(self):
        om = self.om
        config_file = om.defaults.OMEGA_CONFIG_FILE
        if config_file is None:
            config_file = print("No configuration file identified, assuming defaults")
        # print config
        restapi_url = getattr(om.defaults, 'OMEGA_RESTAPI_URL', 'not configured')
        runtime_url = om.runtime.celeryapp.conf['BROKER_URL']
        userid = getattr(om.defaults, 'OMEGA_USERID', 'not configured')
        self.logger.info('Config file: {config_file}'.format(**locals()))
        self.logger.info('User id: {userid}'.format(**locals()))
        self.logger.info('REST API URL: {restapi_url}'.format(**locals()))
        self.logger.info('Runtime broker: {runtime_url}'.format(**locals()))

    def add(self):
        command_url = self._issue_command('install')

    def update(self):
        self._issue_command('update')

    def remove(self):
        self._issue_command('uninstall')

    def _issue_command(self, phase):
        om = self.om
        offering = self.args.get('<kind>')
        size = self.args.get('--count')
        node_type = self.args.get('--node-type')
        user = getattr(om.defaults, 'OMEGA_USERID')
        params = "size={size},node-type={node_type}".format(**locals())
        data = {
            'offering': offering,
            'user': user,
            'phase': phase,
            'params': params,
        }
        command = self._request_service_api(om, 'post', service='command', data=data)
        self._wait_service_command(om, command)

    def _request_service_api(self, om, method, service=None, data=None, uri=None):
        """
        call a service API for a given omega instance

        Args:
            om (Omega): the omega instance, must be authenticated to the cloud
            method (str): the method verb (CREATE, GET, POST, UPDATE, DELETE)
            service (str): the name of the service endpoint as in /api/service/<name>
            data (dict): the data to send

        Returns:
            The response json
        """
        auth = OmegaRestApiAuth.make_from(om)
        restapi_url = getattr(om.defaults, 'OMEGA_RESTAPI_URL', 'not configured')
        uri = uri or '/admin/api/v2/service/{service}'.format(**locals())
        service_url = '{restapi_url}{uri}'.format(**locals())
        method = getattr(requests, method)
        resp = method(service_url, json=data, auth=auth)
        resp.raise_for_status()
        if resp.status_code == requests.codes.created:
            # always request the actual object
            url = resp.headers['Location']
            resp = requests.get(url, auth=auth)
            resp.raise_for_status()
        return resp.json()

    def _wait_service_command(self, om, command):
        """
        wait until the service command status is completed or failed

        Args:
            command (dict): the command object as returned by the service api

        Returns:

        """
        import tqdm
        with tqdm.tqdm(unit='s', total=30) as progress:
            while True:
                progress.update(1)
                uri = command.get('resource_uri')
                data = self._request_service_api(om, 'get', uri=uri)
                status = data.get('status')
                if int(status) > 1:
                    break
                sleep(1)
        if status == '5':
            self.logger.info("Ok, done.")
        else:
            msg = "Error {status} occurred on {uri}".format(**locals())
            self.logger.error(msg)



from unittest import TestCase

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

from omegaml import Omega
from omegaml.client.auth import OmegaRestApiAuth
from omegaml.restapi.app import app
from omegaml.restapi.tests.util import RequestsLikeTestClient
from omegaml.tests.util import OmegaTestMixin


class OmegaRestApiTests(OmegaTestMixin, TestCase):
    def setUp(self):
        self.client = RequestsLikeTestClient(app)
        self.om = Omega()
        self.auth = OmegaRestApiAuth('user', 'pass')
        self.clean()

    @property
    def _headers(self):
        return {}

    def test_predict_from_data_inline(self):
        X = np.arange(10).reshape(-1, 1)
        y = X * 2
        # train model locally
        clf = LinearRegression()
        clf.fit(X, y)
        result = clf.predict(X)
        # store model in om
        self.om.models.put(clf, 'regression')
        # check we can use it to predict
        resp = self.client.put('/api/v1/model/regression/predict', json={
            'columns': ['v'],
            'data': dict(v=[5]),
        }, auth=self.auth, headers=self._headers)
        self.assertEqual(resp.status_code, 200)
        data = resp.get_json()
        self.assertEqual(data.get('model'), 'regression')
        self.assertEqual(data.get('result'), [10.])

    def test_predict_from_dataset(self):
        X = np.arange(10).reshape(-1, 1)
        y = X * 2
        # train model locally
        clf = LinearRegression()
        clf.fit(X, y)
        result = clf.predict(X)
        # store model in om
        self.om.models.put(clf, 'regression')
        self.om.datasets.put([5], 'foo', append=False)
        # check we can use it to predict
        resp = self.client.put('/api/v1/model/regression/predict?datax=foo',
                               json={}, auth = self.auth, headers=self._headers)
        self.assertEqual(resp.status_code, 200)
        data = resp.get_json()
        self.assertEqual(data.get('model'), 'regression')
        self.assertEqual(data.get('result'), [10.])

    def test_dataset_query(self):
        om = self.om
        df = pd.DataFrame({
            'x': np.arange(100),
            'y': np.arange(100),
        })
        om.datasets.put(df, 'test', append=False)
        resp = self.client.get('/api/v1/dataset/test', auth=self.auth, headers=self._headers)
        self.assertEqual(resp.status_code, 200)
        data = resp.get_json()
        self.assertIn('data', data)
        self.assertIn('index', data)
        self.assertIn('values', data['index'])
        self.assertEqual(len(data['index']['values']), 100)
        self.assertEqual(len(data['data']['x']), 100)
        self.assertEqual(len(data['data']['y']), 100)

    def test_dataset_query_filter(self):
        om = self.om
        df = pd.DataFrame({
            'x': np.arange(100),
            'y': np.arange(100),
        })
        om.datasets.put(df, 'test', append=False)
        query = {
            'x__gte': 90,
        }
        resp = self.client.get('/api/v1/dataset/test', query_string=query,
                               auth=self.auth, headers=self._headers)
        self.assertEqual(resp.status_code, 200)
        data = resp.get_json()
        self.assertIn('data', data)
        self.assertIn('index', data)
        self.assertIn('values', data['index'])
        self.assertEqual(len(data['index']['values']), 10)
        self.assertEqual(len(data['data']['x']), 10)
        self.assertEqual(len(data['data']['y']), 10)

    def test_dataset_put(self):
        om = self.om
        om.datasets.drop('foo', force=True)
        # create a new dataset
        data = {
            'data': {
                'x': list(range(10)),
                'y': list(range(10)),
                's': [str(v) for v in range(10)],
            },
            'dtypes': {
                's': 'str',
            },
            'append': False,
        }
        resp = self.client.put('/api/v1/dataset/foo', json=data,
                               auth=self.auth, headers=self._headers)
        self.assertEqual(resp.status_code, 200)
        # -- see if we can query
        df = om.datasets.get('foo')
        self.assertEqual(10, len(df))
        self.assertEqual(list(range(10)), list(df['x']))
        self.assertEqual(list(range(10)), list(df['y']))
        self.assertEqual([str(v) for v in range(10)], list(df['s']))
        # append more records
        data['append'] = True
        resp = self.client.put('/api/v1/dataset/foo', json=data,
                               auth=self.auth, headers=self._headers)
        self.assertEqual(resp.status_code, 200)
        df = om.datasets.get('foo')
        self.assertEqual(20, len(df))
        self.assertEqual(list(range(10)) * 2, list(df['x']))
        self.assertEqual(list(range(10)) * 2, list(df['y']))
        self.assertEqual([str(v) for v in list(range(10)) * 2], list(df['s']))

    def test_dataset_delete(self):
        om = self.om
        # test non-existent dataset
        om.datasets.drop('foo', force=True)
        resp = self.client.delete('/api/v1/dataset/foo',
                                  auth=self.auth, headers=self._headers)
        self.assertEqual(404, resp.status_code)
        df = pd.DataFrame({
            'x': np.arange(100),
            'y': np.arange(100),
        })
        # test existing dataset
        om.datasets.put(df, 'foo', append=False)
        resp = self.client.delete('/api/v1/dataset/foo',
                                  auth=self.auth, headers=self._headers)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(None, om.datasets.get('foo'))

    




import os

import six
import yaml

from omegaml.client.userconf import get_omega_from_apikey
from omegaml.client.util import protected
from omegaml.omega import Omega as CoreOmega
from omegaml.runtimes import OmegaRuntime


class OmegaCloud(CoreOmega):
    """
    Client API to omegaml cloud

    Provides the following APIs:

    * :code:`datasets` - access to datasets stored in the cluster
    * :code:`models` - access to models stored in the cluster
    * :code:`runtimes` - access to the cluster compute resources
    * :code:`jobs` - access to jobs stored and executed in the cluster
    * :code:`scripts` - access to lambda modules stored and executed in the cluster

    """

    def __init__(self, auth=None, **kwargs):
        """
        Initialize the client API
        """
        super(OmegaCloud, self).__init__(**kwargs)
        self.auth = auth

    def _make_runtime(self, celeryconf):
        return OmegaCloudRuntime(self, bucket=self.bucket, defaults=self.defaults, celeryconf=celeryconf)

    def _clone(self, **kwargs):
        kwargs.update(auth=self.auth)
        return super()._clone(**kwargs)

    def __repr__(self):
        return 'OmegaCloud(bucket={}, auth={})'.format(self.bucket or 'default', repr(self.auth))


class OmegaCloudRuntime(OmegaRuntime):
    """
    omegaml hosted compute cluster gateway
    """

    def __init__(self, omega, **kwargs):
        super().__init__(omega, **kwargs)
        self._auth_kwarg = protected('auth')

    def __repr__(self):
        return 'OmegaCloudRuntime(auth={})'.format(repr(self.omega.auth))

    @property
    def _common_kwargs(self):
        common = super()._common_kwargs
        common['task'].update({self._auth_kwarg: self.auth_tuple})
        label = common['routing'].get('label', 'default')
        if getattr(self.omega.defaults, 'OMEGA_TASK_ROUTING_ENABLED', False):
            common['routing'].update(self.build_account_routing(label))
        return common

    @property
    def auth_tuple(self):
        auth = self.omega.auth
        return auth.userid, auth.apikey, auth.qualifier

    @property
    def dispatch_label(self):
        return self.omega.auth.userid

    def build_account_routing(self, label):
        # build the queue
        account_routing = {
            'label': '{self.dispatch_label}-{label}'.format(**locals())
        }
        return account_routing


def setup(userid=None, apikey=None, api_url=None, qualifier=None, bucket=None):
    # from now on, link OmegaCloud implementation as the default
    import omegaml as om
    api_url = api_url or os.environ.get('OMEGA_RESTAPI_URL') or 'https://hub.omegaml.io'
    om.Omega = OmegaCloud
    om.setup = setup
    om.get_omega_for_task = lambda *args, **kwargs: setup(*args, **kwargs)
    om = get_omega_from_apikey(userid, apikey, api_url=api_url, qualifier=qualifier, view=False)
    return om[bucket]


def setup_from_config(config_file=None):
    from omegaml import _base_config
    config_file = config_file or _base_config.OMEGA_CONFIG_FILE
    if isinstance(config_file, six.string_types) and os.path.exists(config_file):
        with open(config_file, 'r') as fin:
            userconfig = yaml.safe_load(fin)
            if isinstance(userconfig, dict) and 'OMEGA_USERID' in userconfig:
                try:
                    omega = setup(userid=userconfig['OMEGA_USERID'],
                                  apikey=userconfig['OMEGA_APIKEY'],
                                  api_url=userconfig['OMEGA_RESTAPI_URL'])
                except Exception as e:
                    # TODO make this a SystemError so that OmegaDeferredIstance.setup reverts to proper defaults
                    raise ValueError('Could not login using config file {}, error={}'.format(config_file, str(e)))
            else:
                _base_config.OMEGA_CONFIG_FILE = config_file
                raise SystemError
            return omega
    raise SystemError('Config file {} does not exist'.format(config_file))



import warnings
from unittest import TestCase

import sys

import os


class OmegaTestMixin(object):
    def clean(self, bucket=None):
        om = self.om[bucket] if bucket is not None else self.om
        for element in ('models', 'jobs', 'datasets', 'scripts'):
            part = getattr(om, element)
            drop = part.drop
            [drop(m.name, force=True) for m in part.list(hidden=True, include_temp=True, raw=True)]
            self.assertListEqual(part.list(hidden=True, include_temp=True), [])

def tf_in_eager_execution():
    # condition for unittest.skipIf decorator
    import tensorflow as tf
    return tf.executing_eagerly()

def tf_perhaps_eager_execution(*args, **kwargs):
    """
    test support to enable tf eager execution

    conditionally set eager execution if TF_EAGER env
    variable is yet to 1. else the eager state is not
    changed.
    """
    tf_eager_switch = os.environ.get('TF_EAGER', False)
    if int(tf_eager_switch):
        import tensorflow as tf
        try:
            tf.enable_eager_execution(*args, **kwargs)
            warnings.warn('TensorFlow eager execution enabled')
        except ValueError as e:
            warnings.warn(str(e))
    else:
        warnings.warn('TensorFlow eager execution not enabled TF_EAGER={tf_eager_switch}'.format(**locals()))

def clear_om(om):
    for omstore in (om.datasets, om.jobs, om.models):
        [omstore.drop(name) for name in omstore.list()]





import datetime
from unittest import TestCase

from omegaml import Omega
from omegaml.backends.virtualobj import VirtualObjectBackend, virtualobj, VirtualObjectHandler
from omegaml.mixins.store.virtualobj import VirtualObjectMixin
from omegaml.tests.util import OmegaTestMixin


class VirtualObjectTests(OmegaTestMixin, TestCase):
    def setUp(self):
        om = self.om = Omega()
        om.datasets.register_backend(VirtualObjectBackend.KIND, VirtualObjectBackend)
        om.datasets.register_mixin(VirtualObjectMixin)
        self.clean()

    def test_put(self):
        om = self.om
        meta = om.datasets.put(myvirtualfn, 'virtualobj')
        self.assertEqual(meta.kind, VirtualObjectBackend.KIND)

    def test_put_get(self):
        om = self.om
        meta = om.datasets.put(myvirtualfn, 'virtualobj')
        self.assertEqual(meta.kind, VirtualObjectBackend.KIND)
        # no data saved through the virtualobj yet
        data = om.datasets.get('virtualobj')
        self.assertEqual(data, 'no data yet')
        # save data
        entrymeta = om.datasets.put(['foo'], 'virtualobj')
        entrymeta = om.datasets.put(['bar'], 'virtualobj')
        self.assertEqual(entrymeta.name, 'virtualobj_data')
        self.assertEqual(entrymeta.attributes['virtualobj_ref'], 'virtualobj')
        meta = om.datasets.metadata('virtualobj')
        self.assertEqual(len(meta.attributes['real_data']), 2)
        data = om.datasets.get('virtualobj')
        self.assertEqual(data, [['foo'], ['bar']])

    def test_put_get_with_basename(self):
        om = self.om
        meta = om.datasets.put(myvirtualobjfn_with_basename, 'virtualobj')
        self.assertEqual(meta.kind, VirtualObjectBackend.KIND)
        # no data saved through the virtualobj yet
        data = om.datasets.get('virtualobj{base_name=foo_data}')
        self.assertEqual(data, 'no data yet')
        # save data
        entrymeta = om.datasets.put(['foo'], 'virtualobj{base_name=foo_data}')
        entrymeta = om.datasets.put(['bar'], 'virtualobj{base_name=foo_data}')
        self.assertEqual(entrymeta.name, 'foo_data')
        self.assertEqual(entrymeta.attributes['virtualobj_ref'], 'virtualobj')
        meta = om.datasets.metadata('virtualobj')
        self.assertEqual(len(meta.attributes['real_data']), 2)
        data = om.datasets.get('virtualobj{base_name=foo_data}')
        self.assertEqual(data, [['foo'], ['bar']])

    def test_drop(self):
        om = self.om
        meta = om.datasets.put(myvirtualfn, 'virtualobj')
        self.assertEqual(meta.kind, VirtualObjectBackend.KIND)

    def test_virtualhandler(self):
        om = self.om
        om.datasets.put(MyVirtualObjectHandler, 'virtualobj')
        entrymeta = om.datasets.put(['foo'], 'virtualobj')
        entrymeta = om.datasets.put(['bar'], 'virtualobj')
        self.assertEqual(entrymeta.name, 'virtualobj_data')

@virtualobj
def myvirtualfn(data=None, meta=None, method=None, store=None, **kwargs):
    real_data_name = '{}_data'.format(meta.name)
    if method == 'get':
        data = store.get(real_data_name)
        return data or 'no data yet'
    if method == 'put':
        entrymeta = store.put(data, real_data_name, attributes={
            'virtualobj_ref': meta.name,
        })
        entrylist = meta.attributes.get('real_data', [])
        entrylist.append(datetime.datetime.now())
        meta.attributes['real_data'] = entrylist
        meta.save()
        return entrymeta
    if method == 'drop':
        store.drop(real_data_name, force=True)
        return 'ok, deleted'
    raise NotImplementedError


class MyVirtualObjectHandler(VirtualObjectHandler):
    def real_data_name(self, meta):
        return '{}_data'.format(meta.name)

    def get(self, data=None, meta=None, store=None, **kwargs):
        return store.get(self.real_data_name(meta)) or 'no data yet'

    def put(self, data=None, meta=None, store=None, **kwargs):
        entrymeta = store.put(data, self.real_data_name(meta), attributes={
            'virtualobj_ref': meta.name,
        })
        entrylist = meta.attributes.get('real_data', [])
        entrylist.append(datetime.datetime.now())
        meta.attributes['real_data'] = entrylist
        meta.save()
        return entrymeta

    def drop(self, data=None, meta=None, store=None, **kwargs):
        store.drop(self.real_data_name(meta), force=True)
        return 'ok, deleted'

@virtualobj
def myvirtualobjfn_with_basename(data=None, meta=None, method=None, base_name=None, store=None, **kwargs):
    real_data_name = base_name
    if method == 'get':
        data = store.get(real_data_name)
        return data or 'no data yet'
    if method == 'put':
        entrymeta = store.put(data, real_data_name, attributes={
            'virtualobj_ref': meta.name,
        })
        entrylist = meta.attributes.get('real_data', [])
        entrylist.append(datetime.datetime.now())
        meta.attributes['real_data'] = entrylist
        meta.save()
        return entrymeta
    if method == 'drop':
        store.drop(real_data_name, force=True)
        return 'ok, deleted'
    raise NotImplementedError


import copy

import six

from omegaml.store.queryops import MongoQueryOps, flatten_keys
from omegaml.util import restore_index


class MongoQ(object):

    """
    Query object to filter mongodb collections

    A Q object holds the conditions. You can combine Q objects
    using & and | operators, e.g.

    q1 = Q(year=2015)
    q2 = Q(year=2014)
    q_all = q1 | q2

    Query objects are passed into a Filter, and evaluated when
    the filter's .value property is read. 

    Conditions specified in one Q object are AND together. Various
    operators can be specified using __<op> syntax, e.g. year__eq=2015.

    eq     ==  (default)
    lt     <
    lte    <=
    gt     >
    gte    >=
    ne     <>
    not    <>
    in     isin(v) 
    between

    startswith
    endwith
    contains  
    match     

    Evaluation works as follows:

    1. for each Q object, apply the filters as collection.find(query)
    2. return the dataframe 

    """
    def __init__(self, **kwargs):
        self.conditions = kwargs
        self.qlist = [('', self)]
        # should we return ~(conditions)
        self._inv = False
        # is sorting implied by some operator
        self.sorted = False

    def __repr__(self):
        r = []
        for op, q in self.qlist:
            r.append('%s %s' % (op, q.conditions))
        return 'Q %s' % ('\n'.join(r))

    def value(self, collection):
        """
        resolve the query by actually applying the filter on given collection

        :param collection: the collection objec
        :return: the result of the apply_filter() call 
        """
        return self.apply_filter(collection)

    def apply_filter(self, collection):
        """
        apply the filter to the given collection

        This builds the actual mongo query using build_filters and
        calls apply_conditions to execute the query 

        :param collection: the collection object
        :return: the result of the apply_conditions() call
        """
        query = self.build_filters()
        return self.apply_conditions(collection, query)

    def apply_conditions(self, collection, query):
        """
        apply the mongo query on a given collection

        :param collection: the collection 
        :param query: the query dictionary applicable to collection.find()
        :return: the result of collection.find()  
        """
        operators = flatten_keys(query)
        if '$near' in operators:
            self.sorted = True
        return collection.find(query)

    def build_filters(self):
        """
        build the complex mongodb filter from query definitions

        This takes all query definitions in qlist, builds queries
        using .build_conditions() and concatenates the queries by
        and/or as defined the in qlist's op value.

        :return: the query suitable for the collection.find() method
        """
        query = {}
        for i, (op, q) in enumerate(self.qlist):
            if i == 0:
                query = q.build_conditions()
            elif op == '&':
                query.setdefault("$and", [])
                fn = q.build_conditions if q == self else q.build_filters
                query["$and"].append(fn())
            elif op == '|':
                fn = q.build_conditions if q == self else q.build_filters
                if i == 1:
                    query = {"$or": [dict(query)]}
                else:
                    query.setdefault("$or", [])
                query["$or"].append(fn())
        if self._inv:
            query = {"$nor": [query]}
        return query

    def build_conditions(self):
        """
        For a given query definition return the collection.find() simple query

        Using all conditions, build the query as a dictionary suitable for
        collection.find(). This uses MongoQueryOps to transform query 
        definitions into mongo db syntax.

        :return: the query in mongo db syntax  
        """
        query = {}
        qops = MongoQueryOps()
        def addq(k, v):
            if k not in query:
                query[k] = v
            else:
                subq = []
                query.setdefault("$and", subq)
                for vv in [query.pop(k), v]:
                    if isinstance(vv, (list, tuple)):
                        subq.extend(vv)
                    else:
                        subq.append({k: vv})
        for k, v in six.iteritems(self.conditions):
            # transform query operators as '<foo>__<op>',
            # however preserve dunder '__<foo>' names ss columns
            if '__' in k and not k.startswith('__'):
                parts = k.split('__')
                k = '.'.join(parts[0:-1])
                op = parts[-1]
            else:
                op = 'eq'
            # standard logical operators
            if op == 'eq':
                addq(k, v)
            elif op.upper() in qops.UNARY:
                addq(k, getattr(qops, op)(v))
            # type queries
            elif op == 'between':
                addq("$and", [{k: qops.GTE(v[0])},
                              {k: qops.LTE(v[1])}])
            elif op == 'isstring':
                addq(k, qops.EQ(qops.TYPE('string')))
            elif op == 'isarray':
                addq(k, qops.EQ(qops.TYPE('array')))
            elif op == 'isdouble':
                addq(k, qops.TYPE('double'))
            elif op == 'isobject':
                addq(k, qops.TYPE('object'))
            elif op == 'isobject':
                addq(k, qops.TYPE('object'))
            elif op == 'isdate':
                addq(k, qops.TYPE('date'))
            elif op == 'isbool':
                addq(k, qops.TYPE('bool'))
            elif op == 'isnull':
                # http://stackoverflow.com/a/944733
                nan = float('nan')
                addq(k, qops.EQ(nan) if v else qops.NE(nan))
            elif op in ['islong', 'isint']:
                addq(k, qops.TYPE('long'))
            elif op == 'contains':
                addq(k, qops.REGEX('^%s.*' % v))
            elif op == 'startswith':
                addq(k, qops.REGEX('^%s.*' % v))
            elif op == 'endswith':
                addq(k, qops.REGEX('.*%s$' % v))
            elif op == 'near':
                addq(k, qops.NEAR(v))
            else:
                # op from parts[-1] was not an opperator, so assume it is
                # an attribute name and apply the eq operator
                # e.g. Q(key__subkey=value)
                addq('%s.%s' % (k, op), v)
        return query

    def negate(self):
        """
        negate the query
        """
        self._inv = True
        return self

    def __and__(self, other):
        """
        combine with another MongoQ object using AND
        """
        q = copy.deepcopy(self)
        q.qlist.append(('&', other))
        return q

    def __or__(self, other):
        """
        combine with another MongoQ object using OR
        """
        q = copy.deepcopy(self)
        q.qlist.append(('|', other))
        return q

    def __invert__(self):
        """
        return an inverted version of this object
        """
        notq = copy.deepcopy(self).negate()
        return notq


class Filter(object):

    """
    Filter for OmegaStore objects

    Allow keyword style filtering of dataframes:

    direct filtering
        filter = Filter(df, year=2015)
        filter.value

    filtering with Q objects
        q = Q(year=2015)
        filter = Filter(df, q)

    filtering with multiple Q objects
        q1 = Q(year=2015)
        q2 = Q(year=2014)
        filter = Filter(df, q1 | q2)

    filtering with filter() and exclude()
        filter = Filter(df, year=2015)
        filter.filter(month=1)
        filter.exclude(day=15)
    """
    _debug = False

    def __init__(self, coll, __query=None, **kwargs):
        """
        Filter objects use MongoQ query expression to build
        complex filters.

        Internally, Filter keeps the .q MongoQ object to represent
        the filter. Filter thus is a thin wrapper around MongoQ to 
        represent the human-readable high-level API to MongoQ:

        .filter -- add a new condition with AND
        .exclude -- exclude values given a condition 
        .count -- count result lengths. note this triggers .value
        .value -- evaluate the filter

        In addition Filter supports tracing errors by using the .trace
        flag (set to True, defaults to False).

        :param coll: the collection to apply this filter to
        :param __query: a MongoQ object (used internally to combine
        filters)
        :param kwargs: the filter as a dict of column__op=value pairs
        """
        self.coll = coll
        self.trace = False
        self.exc = None
        self.q = self.build_query(__query, **kwargs)

    def build_query(self, __query=None, **kwargs):
        """
        build the MongoQ object from given kwargs

        Specify either the __query or the kwargs. If __query is
        specified it is returned unchanged. If kwargs is passed these
        are used to build a new MongoQ object. The rationale for this
        is so that build_query can be used without an if statement in
        the calling code (i.e. this methods wraps the if statement on
        whether to build a new MongoQ object or to use the existing,
        making for more readable code).

        :param __query: the MongoQ object to use
        :param kwargs: the filter as a dict of column__op=value pairs
        :return: the MongoQ object
        """
        if __query:
            q = __query
        else:
            q = MongoQ(**kwargs)
        return q

    @property
    def query(self):
        """
        Convenience method to return the filter's query in MongoDB syntax
        """
        return self.q.build_filters()

    def count(self):
        """
        Resolve the query and count number of rows
        """
        return len(self.value.index)

    @property
    def value(self):
        """
        Resolve the query and return its results

        Uses self.evaluate() in a safe manner.
        """
        try:
            value = self.evaluate()
        except KeyError as e:
            self.exc = e
            if self.trace:
                raise
            raise SyntaxError(
                'Error in Q object: column %s is unknown (KeyError on dataframe)' % e)
        return value

    def filter(self, query=None, **kwargs):
        """
        Add a new query expression using AND

        :param query: an existing MongoQ object (optional)
        :param kwargs: the kwargs as column__op=value pairs
        """
        self.q &= self.build_query(query, **kwargs)
        return self

    def exclude(self, query=None, **kwargs):
        """
        Add a new negated query expression

        This is the equivalent of .filter(~query)

        :param query: an existing MongoQ object (optional)
        :param kwargs: the kwargs as column__op=value pairs
        """
        self.q &= ~self.build_query(query, **kwargs)
        return self

    def __invert__(self):
        return Filter(self.coll, ~self.q)

    def __and__(self, other):
        return Filter(self.coll, self.q & other.q)

    def __or__(self, other):
        return Filter(self.coll, self.q | other.q)

    def evaluate(self):
        """
        evaluate the query

        :return: the pandas DataFrame
        """
        result = self.q.apply_filter(self.coll)
        try:
            import pandas as pd
            result = pd.DataFrame.from_records(result)
            if '_id' in result.columns:
                del result['_id']
            result = restore_index(result, dict(),
                                   rowid_sort=not self.q.sorted)
        except ImportError:
            result = list(result)
        return result

    def as_mask(self):
        """
        return the True mask of this filter

        note this does not actually return a mask but applies the filter
             to the real data. makes more sense even though the results
             are somewhat different this way, but match better semantically
        """
        return self.value == self.value

    def _repr_html_(self):
        return self.value._repr_html_()

    def __repr__(self):
        return self.value.__repr__()



from __future__ import absolute_import
import json
import sys

import pymongo

from omegaml.util import make_tuple
import six


class GeoJSON(dict):

    """
    simple GeoJSON object

    input:
        GeoJSON(lon, lat)
        GeoJSON('lon,lat')
        GeoJSON(geojson-object)
        GeoJSON({ geojson dict with 'coordinates': [lon, lat] })
        GeoJSON(coordinates=[lon, lat])

    output:
        GeoJSON.to_dict()
        GeOJSON.to_json()  
    """
    def __init__(self, lon=None, lat=None, coordinates=None):
        if isinstance(lon, GeoJSON):
            coordinates = [lon.lon, lon.lat]
        elif isinstance(lon, (float, int)) and isinstance(lat, (float, int)):
            coordinates = [float(lon), float(lat)]
        elif isinstance(lon, dict):
            coordinates = self.get_coordinates_from_geojson(lon)
        elif isinstance(lon, (list, tuple)):
            coordinates = lon
        elif isinstance(lon, six.string_types):
            coordinates = [float(c) for c in lon.split(',')]
        elif isinstance(coordinates, GeoJSON):
            coordinates = [coordinates.lon, coordinates.lat]
        elif isinstance(coordinates, (list, tuple)):
            coordinates = coordinates
        elif isinstance(coordinates, dict):
            coordinates = self.get_coordinates_from_geojson(lon)
        elif isinstance(coordinates, six.string_types):
            coordinates = [float(c) for c in lon.split(',')]
        else:
            coordinates = []
        self.update(self.to_dict(coordinates))
        assert coordinates, "%s is not a valid coordinate" % coordinates
    def get_coordinates_from_geojson(self, d):
        if 'coordinates' in d:
            coordinates = d.get('coordinates')
        elif 'geometry' in d \
                and d.get('geometry').get('type') == 'Point':
            coordinates = d.get('geometry').get('coordinates')
        else:
            raise ValueError(
                'expected a valid GeoJSON dict, got %s' % coordinates)
        return coordinates
    @property
    def lat(self):
        return self.get('coordinates')[1]
    @property
    def lon(self):
        return self.get('coordinates')[0]
    def to_dict(self, coordinates=None):
        return {
            'type': 'Point',
            'coordinates': coordinates or self.get('coordinates'),
        }
    def to_json(self):
        return json.dumps(self.to_dict())
    def __unicode__(self):
        return u"%s" % self.to_json()


class MongoQueryOps(object):

    """
    A Pythonic API to build Mongo query statements 

    Examples:

    # setup
    x = MongoQueryOps()
    d = dict

    # build queries
    query = x.MATCH(d(foo='value', baz=x.CONTAINS('value'))) 
    groupby = x.GROUP(columns=[col1, col2]), count=x.COUNT())

    result = coll.find(query)
    result = coll.aggregate([query, groupby])
    """

    UNARY = ('IN,LT,LTE,GT,GTE,NE,WHERE,GEOWITHIN,ALL,ELEMWITHIN,NIN'
             'EXISTS,TYPE,REGEX,EQ').split(',')

    def __getattr__(self, k):
        if k.upper().replace('_', '') in MongoQueryOps.UNARY:
            return self.__unary(k.lower())
        raise AttributeError('operator %s is not supported' % k)
    def __unary(self, op):
        """
        return a function to create unary operators

        e.g. MongoQueryOps().lt(val) will return an unary function
        that on being called will return { "$lt" : val }
        """
        def unary(val):
            return {"$%s" % op.lower(): val}
        return unary
    def OR(self, sub):
        return {"$or": sub}
    def AND(self, sub):
        return {"$and": sub}
    def NOT(self, sub):
        return {"$not": sub}
    def GROUP(self, v=None, columns=None, **kwargs):
        from collections import OrderedDict
        if not v:
            v = OrderedDict()
        if not columns:
            v['_id'] = None
        else:
            v.setdefault('_id', {})
            v['_id'].update({k: '$%s' % (k.replace('__', '.'))
                             for k in columns})
        if kwargs:
            for _k, _v in six.iteritems(kwargs):
                v.setdefault(_k, {})
                if isinstance(_v, dict):
                    v[_k].update(_v)
                else:
                    v[_k] = _v
        return {"$group": v}
    def SUM(self, v):
        return {"$sum": v}
    def COUNT(self):
        return self.SUM(1)
    def IS(self, **kwargs):
        return kwargs
    def as_dataframe(self, result, autoflat=True, flatten=None, groupby=None):
        """ transform a resultset into a dataframe"""
        import pandas as pd
        def do_flatten(seq):
            """ extract composed keys into columns """
            for r in seq:
                row = {}
                row.update(r)
                if flatten in list(row.keys()):
                    row.update(row.get(flatten))
                yield row
        if autoflat or flatten == True:
            flatten = '_id'
        df = pd.DataFrame(do_flatten(result))
        if groupby and len(df.index) > 0:
            if isinstance(groupby, bool):
                cols = list(df.iloc[0]['_id'].keys())
            else:
                cols = groupby
            df.set_index(cols, inplace=True)
        return df
    def MATCH(self, *args, **kwargs):
        if args:
            v = args[0]
        else:
            v = kwargs
        return {"$match": v}
    def SEARCH(self, v):
        return {"$text": {"$search": v}}
    def CONTAINS(self, v):
        return {"$regex": '.*%s.*' % v}
    def SORT(self, **columns):
        """
        sort by columns
        """
        return {"$sort": columns}
    def d(self, **kwargs):
        return dict(**kwargs)
    def to_latex(self, df, fout=None):
        fout = fout or sys.stdout
        fout.write(df.to_latex())
        return fout
    def PROJECT(self, fields, include=True):
        fields = make_tuple(fields)
        return {
            '$project': { key: 1 if include else 0 for key in fields}
        }
    def LOOKUP(self, other, key=None, left_key=None, right_key=None,
               target=None):
        """
        return a $lookup statement.

        :param other: the other collection
        :param key: the key field (applies to both left and right)
        :param left_key: the left key field
        :param right_key: the right key field 
        :param target: the target array to store the matching other-documents
        """
        return {
            "$lookup": {
                "from": other,
                "localField": left_key or key,
                "foreignField": right_key or key,
                "as": target or ("%s_%s" % (other, key or right_key))
            }
        }
    def UNWIND(self, field, preserve=True, index=None):
        """
        returns $unwind for the given array field. the index in the
        array will be output as _index_<field>. 

        :param field: the array field to unwind from
        :param preserve: if True, the document is output even if the
           array field is empty.
        :param index: if given the index field is taken from this field
        """
        op = {
            "$unwind": {
                "path": "${}".format(field)
            }
        }
        if preserve is not None:
            op['$unwind'].update({
                "preserveNullAndEmptyArrays": preserve
            })
        if index is not None:
            op['$unwind'].update({
                "includeArrayIndex": "%s_%s" % ('_index_', index),
            })
        return op
    def OUT(self, name):
        return {"$out": name}
    def SET(self, column, value):
        return {"$set": {column: value}}
    def NEAR(self, lon=None, lat=None, location=None, maxd=None, mind=None):
        """
        return a $near expression from an explicit lon/lat coordinate, a
        GeoJSON object, a GeoJSON dictionary or a string
        """
        if isinstance(lon, six.string_types):
            location = GeoJSON(lon)
        elif isinstance(lon, (list, tuple)):
            if len(lon) == 4:
                lon, lat, mind, maxd = lon
            elif len(lon) == 3:
                lon, lat, maxd = lon
            else:
                lon, lat = lon
            location = GeoJSON(lon, lat)
        elif isinstance(lon, GeoJSON):
            location = lon
        elif isinstance(lon, dict):
            location = GeoJSON(lon.get('location'))
            maxd = lon.get('maxd')
            mind = lon.get('mind')
        elif not location:
            assert "invalid arguments. Specify coordinates=GeoJSON(lon, lat)"
        else:
            pass
        if isinstance(location, (list, tuple)):
            lon, lat = location
        else:
            lon, lat = location.get('coordinates')
        assert lon, "invalid coordinate lon=%s lat=%s" % (lon, lat)
        assert lat, "invalid coordinate lon=%s lat=%s" % (lon, lat)
        nearq = {
            '$near': {
                '$geometry': {
                    'type': 'Point',
                    'coordinates': [lon, lat],
                },
            }
        }
        if maxd:
            nearq['$near']['$maxDistance'] = maxd
        if mind:
            nearq['$near']['$minDistance'] = mind
        return nearq
    def REPLACEROOT(self, field):
        return {
            '$replaceRoot': {
                'newRoot': "${}".format(field)
            }
        }
    def make_index(self, columns, **kwargs):
        """
        return an index specification suitable for collection.create_index()

        using columns specs like ['+A', '-A'] returns (key, index)
        pairs suitable for passing on to create_index. also generates
        a name for the index based on the columns and ordering. Use
        '@coord' to create a geospecial index. The coord column must
        be in GeoJSON format

        :param columns: a single index column, or a list of columns
        :param kwargs: optional kwargs to merge. if kwargs contains the
        'name' key it will be preserved
        :return: (idx, **kwargs) tuple, pass as create_index(idx, **kwargs)  
        """
        SORTPREFIX = ['-', '+', '@']
        DIRECTIONMAP = {
            '-': pymongo.DESCENDING,
            '+': pymongo.ASCENDING,
            '@': pymongo.GEOSPHERE,
            'default': pymongo.ASCENDING,
        }
        columns = make_tuple(columns)
        direction_default = DIRECTIONMAP.get('default')
        sort_cols = ['+' + col
                     if col[0] not in SORTPREFIX else col for col in columns]
        # get sort kwargs
        def direction(col):
            return DIRECTIONMAP.get(col[0], direction_default)
        idx = [(col.replace('+', '').replace('-', '').replace('@', ''),
                direction(col))
               for col in sort_cols]
        name = '__'.join([(col
                           .replace('-', 'desc_')
                           .replace('+', 'asc_')
                           .replace('@', 'geo_'))
                          for col in sort_cols])
        kwargs.setdefault('name', name)
        return idx, kwargs
    def make_sortkey(self, columns):
        """
        using columns specs like ['+A', '-A'] returns (key, index)
        pairs suitable for passing on to collection.sort()
        """
        sort_key, _ = self.make_index(columns)
        return sort_key


def flatten_keys(d, keys=None):
    """
    get all keys from a dictionary and nested dicts, as a flattened list

    :param d: a dictionary
    :param keys: previously found keys. internal use.
    :returns: list of flattened keys 
    """
    keys = keys or []
    keys.extend(list(d.keys()))
    for sd in d.values():
        if isinstance(sd, dict):
            flatten_keys(sd, keys=keys)
    return keys


# convenience accessors
x = MongoQueryOps()
d = dict



from __future__ import absolute_import

import unittest
import uuid
from unittest import skip
from zipfile import ZipFile

import gridfs
import pandas as pd
from datetime import timedelta
from mongoengine.connection import disconnect
from mongoengine.errors import DoesNotExist, FieldDoesNotExist
from pandas.io.json import json_normalize
from pandas.util import testing
from pandas.util.testing import assert_frame_equal, assert_series_equal
from six import BytesIO
from six.moves import range
from sklearn.datasets import load_iris
from sklearn.linear_model import LogisticRegression

from omegaml import backends
from omegaml.backends.rawdict import PandasRawDictBackend
from omegaml.backends.rawfiles import PythonRawFileBackend
from omegaml.documents import MDREGISTRY
from omegaml.mdataframe import MDataFrame
from omegaml.store import OmegaStore
from omegaml.util import delete_database


class StoreTests(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        delete_database()

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        delete_database()
        disconnect('omega')

    def test_package_model(self):
        # create a test model
        store = OmegaStore()
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        result = lr.predict(X)
        # package locally
        backend = backends.ScikitLearnBackend(model_store=store,
                                              data_store=store)
        zipfname = backend._package_model(lr, 'models/foo')
        # load it, try predicting
        lr2 = backend._extract_model(zipfname)
        self.assertIsInstance(lr2, LogisticRegression)
        result2 = lr2.predict(X)
        self.assertTrue((result == result2).all())

    def test_put_model(self):
        # create a test model
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        result = lr.predict(X)
        # store it remote
        store = OmegaStore()
        store.put(lr, 'models/foo')
        # get it back, try predicting
        lr2 = store.get('models/foo')
        self.assertIsInstance(lr2, LogisticRegression)
        result2 = lr2.predict(X)
        self.assertTrue((result == result2).all())

    def test_prefix_store(self):
        """
        this is to test if store prefixes work
        """
        df = pd.DataFrame({
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        })
        datasets = OmegaStore(prefix='teststore')
        models = OmegaStore(prefix='models', kind=MDREGISTRY.SKLEARN_JOBLIB)
        datasets.put(df, 'test')
        self.assertEqual(len(datasets.list()), 1)
        self.assertEqual(len(models.list()), 0)

    def test_custom_levels(self):
        """
        this is to test if custom path and levels can be provided ok
        """
        df = pd.DataFrame({
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        })
        datasets = OmegaStore(prefix='data')
        models = OmegaStore(prefix='models', kind=MDREGISTRY.SKLEARN_JOBLIB)
        # directory-like levels
        datasets.put(df, 'data/is/mypath/test')
        datasets.put(df, 'data/is/mypath/test2')
        self.assertEqual(len(datasets.list('data/*/mypath/*')), 2)
        self.assertEqual(len(datasets.list('data/*/test')), 1)
        # namespace-like levels
        datasets.put(df, 'my.namespace.module.test')
        datasets.put(df, 'my.namespace.module.test2')
        self.assertEqual(len(datasets.list('*.module.*')), 2)
        self.assertEqual(len(datasets.list('*.module.test2')), 1)

    def test_put_model_with_prefix(self):
        # create a test model
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        result = lr.predict(X)
        # store it remote
        store = OmegaStore(prefix='models/')
        store.put(lr, 'foo')
        # get it back, try predicting
        lr2 = store.get('foo')
        self.assertIsInstance(lr2, LogisticRegression)
        result2 = lr2.predict(X)
        self.assertTrue((result == result2).all())

    def test_put_dataframe(self):
        # create some dataframe
        df = pd.DataFrame({
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        })
        store = OmegaStore(prefix='')
        store.put(df, 'mydata')
        df2 = store.get('mydata')
        self.assertTrue(df.equals(df2), "expected dataframes to be equal")

    def test_put_dataframe_xtra_large(self):
        # create some dataframe
        # force fast insert
        df = pd.DataFrame({
            'a': list(range(0, int(1e4 + 1))),
            'b': list(range(0, int(1e4 + 1)))
        })
        store = OmegaStore(prefix='')
        store.put(df, 'mydata')
        df2 = store.get('mydata')
        self.assertTrue(df.equals(df2), "expected dataframes to be equal")

    def test_put_dataframe_timestamp(self):
        # create some dataframe
        from datetime import datetime
        df = pd.DataFrame({
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        })
        store = OmegaStore(prefix='')
        # -- check default timestamp
        now = datetime.utcnow()
        store.put(df, 'mydata', append=False, timestamp=True)
        df2 = store.get('mydata')
        _created = pd.to_datetime(df2['_created'].unique()[0])
        self.assertEqual(_created.replace(second=0, microsecond=0),
                         now.replace(second=0, microsecond=0))
        # -- check custom timestamp column, default value
        now = datetime.utcnow()
        store.put(df, 'mydata', append=False, timestamp='CREATED')
        df2 = store.get('mydata')
        _created = pd.to_datetime(df2['CREATED'].unique()[0])
        self.assertEqual(_created.replace(second=0, microsecond=0),
                         now.replace(second=0, microsecond=0))
        # -- check custom timestamp column, value as tuple
        now = datetime.utcnow() - timedelta(days=1)
        store.put(df, 'mydata', append=False, timestamp=('CREATED', now))
        df2 = store.get('mydata')
        _created = pd.to_datetime(df2['CREATED'].unique()[0])
        self.assertEqual(_created.replace(second=0, microsecond=0),
                         now.replace(second=0, microsecond=0))
        # set a day in the past to avoid accidentally creating the current
        # datetime in mongo
        now = datetime.now() - timedelta(days=1)
        store.put(df, 'mydata', timestamp=now, append=False)
        df2 = store.get('mydata')
        # compare the data
        _created = pd.to_datetime(df2['_created'].unique()[0])
        self.assertEqual(_created.replace(microsecond=0),
                         now.replace(microsecond=0))

    def test_get_dataframe_filter(self):
        # create some dataframe
        df = pd.DataFrame({
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        })
        store = OmegaStore(prefix='')
        store.put(df, 'mydata')
        # filter in mongodb
        df2 = store.get('mydata', filter=dict(a__gt=1, a__lt=10))
        # filter local dataframe
        df = df[(df.a > 1) & (df.a < 10)]
        self.assertTrue(df.equals(df2), "expected dataframes to be equal")

    def test_get_dataframe_project(self):
        # create some dataframe
        df = pd.DataFrame({
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        })
        store = OmegaStore(prefix='')
        store.put(df, 'mydata')
        # filter in mongodb
        df2 = store.get('mydata', columns=['a'])
        # filter local dataframe
        df = df[['a']]
        self.assertTrue(df.equals(df2), "expected dataframes to be equal")

    def test_get_dataframe_projected_mixin(self):
        # create some dataframe
        df = pd.DataFrame({
            'a': list(range(1, 10)),
            'b': list(range(1, 10)),
            'c': list(range(1, 10)),
        })
        store = OmegaStore(prefix='')
        store.put(df, 'mydata')
        # filter in mongodb
        specs = ['a', ':b', ':', 'b:', '^c']
        for spec in specs:
            name_spec = 'mydata[{}]'.format(spec)
            df2 = store.get(name_spec)
            # filter local dataframe
            if spec == ':':
                dfx = df.loc[:, :]
            elif ':' in spec:
                from_col, to_col = spec.split(':')
                slice_ = slice(from_col or None, to_col or None)
                dfx = df.loc[:, slice_]
            elif spec.startswith('^'):
                spec_cols = spec[1:].split(',')
                cols = [col for col in df.columns if col not in spec_cols]
                dfx = df[cols]
            else:
                dfx = df[[spec]]
            self.assertTrue(dfx.equals(df2), "expected dataframes to be equal")

    def test_put_dataframe_with_index(self):
        # create some dataframe
        df = pd.DataFrame({
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        })
        store = OmegaStore(prefix='')
        store.put(df, 'mydata', index=['a', '-b'])
        idxs = list(store.collection('mydata').list_indexes())
        idx_names = map(lambda v: dict(v).get('name'), idxs)
        self.assertIn('asc_a__desc_b', idx_names)

    def test_put_dataframe_timeseries(self):
        # create some dataframe
        tsidx = pd.date_range(pd.datetime(2016, 1, 1), pd.datetime(2016, 4, 1))
        df = pd.DataFrame({
            'a': list(range(0, len(tsidx))),
            'b': list(range(0, len(tsidx)))
        }, index=tsidx)
        store = OmegaStore(prefix='')
        store.put(df, 'mydata')
        dfx = store.get('mydata')
        assert_frame_equal(df, dfx)
        idxs = list(store.collection('mydata').list_indexes())
        idx_names = [dict(v).get('name') for v in idxs]
        self.assertIn('asc__idx#0_0', idx_names)

    def test_put_dataframe_multiindex(self):
        # create some dataframe
        store = OmegaStore(prefix='')
        midx = pd.MultiIndex(levels=[[u'bar', u'baz', u'foo', u'qux'],
                                     [u'one', u'two']],
                             codes=[
                                 [0, 0, 1, 1, 2, 2, 3, 3],
                                 [0, 1, 0, 1, 0, 1, 0, 1]],
                             names=[u'first', u'second'])
        df = pd.DataFrame({'x': range(0, len(midx))}, index=midx)
        store.put(df, 'mydata')
        dfx = store.get('mydata')
        assert_frame_equal(df, dfx)
        idxs = list(store.collection('mydata').list_indexes())
        idx_names = [dict(v).get('name') for v in idxs]
        self.assertIn('asc__idx#0_first__asc__idx#1_second', idx_names)

    def test_put_python_dict(self):
        # create some data
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        store = OmegaStore(prefix='')
        store.put(data, 'mydata')
        data2 = store.get('mydata')
        self.assertEquals([data], data2)

    def test_put_python_dict_multiple(self):
        # create some data
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        store = OmegaStore(prefix='')
        store.put(data, 'mydata')
        store.put(data, 'mydata')
        data2 = store.get('mydata')
        # we will have stored the same object twice
        self.assertEquals(data, data2[0])
        self.assertEquals(data, data2[1])

    def test_get_forced_python(self):
        """
        this tests we can retrieve data as python values

        the purpose is to test the basic mode of OmegaStore in
        case pandas and scikit learn are not available
        """
        store = OmegaStore(prefix='')
        # pure data
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        meta = store.put(data, 'data')
        data2 = store.get('data', force_python=True)
        self.assertEqual(data, data2)
        # dataframe
        # create some dataframe
        df = pd.DataFrame({
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        })
        store.put(df, 'mydata')
        df2 = store.get('mydata', force_python=True)
        df2 = pd.DataFrame(df2)
        real_cols = [col for col in df2.columns
                     if (col != '_id'
                         and not col.startswith('_idx')
                         and not col.startswith('_om'))]
        df2 = df2[real_cols]
        self.assertTrue(df.equals(df2), "expected dataframes to be equal")
        # model
        iris = load_iris()
        X = iris.data
        Y = iris.target
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        lr.fit(X, Y)
        # store it remote
        store.put(lr, 'foo')
        # get it back as a zipfile
        lr2file = store.get('foo', force_python=True)
        contents = lr2file.read()
        with ZipFile(BytesIO(contents)) as zipf:
            self.assertIn('foo', zipf.namelist())

    def test_store_with_metadata(self):
        om = OmegaStore(prefix='')
        # dict
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        attributes = {'foo': 'bar'}
        meta = om.put(data, 'data', attributes=attributes)
        self.assertEqual(meta.kind, 'python.data')
        self.assertEqual(meta.attributes, attributes)
        data2 = om.get('data')
        self.assertEqual([data], data2)
        # dataframe
        df = pd.DataFrame(data)
        meta = om.put(df, 'datadf', attributes=attributes)
        self.assertEqual(meta.kind, 'pandas.dfrows')
        self.assertEqual(meta.attributes, attributes)
        df2 = om.get('datadf')
        assert_frame_equal(df, df2)
        # model
        lr = LogisticRegression(solver='liblinear', multi_class='auto')
        meta = om.put(lr, 'mymodel', attributes=attributes)
        self.assertEqual(meta.kind, 'sklearn.joblib')
        self.assertEqual(meta.attributes, attributes)
        lr2 = om.get('mymodel')
        self.assertIsInstance(lr2, LogisticRegression)

    def test_store_metadata_notstrict(self):
        """ ensure Metadata attributes are not strictly checked

        this is to allow metadata extensions between omegaml versions
        """
        om = OmegaStore(prefix='')
        # dict
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        attributes = {'foo': 'bar'}
        meta = om.put(data, 'data', attributes=attributes)
        meta_collection = om.mongodb['metadata']
        flt = {'name': 'data'}
        meta_entry = meta_collection.find_one(flt)
        meta_entry['modified_extra'] = meta_entry['modified']
        meta_collection.replace_one(flt, meta_entry)
        try:
            meta = om.metadata('data')
        except FieldDoesNotExist:
            not_raised = False
        else:
            not_raised = True
        self.assertTrue(not_raised)

    def test_store_dataframe_as_dfgroup(self):
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        result_data = {
            'a': list(range(1, 2)),
            'b': 1,
        }
        df = pd.DataFrame(data)
        result_df = pd.DataFrame(result_data)
        store = OmegaStore()
        groupby_columns = ['b']
        meta = store.put(df, 'dfgroup', groupby=groupby_columns)
        self.assertEqual(meta.kind, 'pandas.dfgroup')
        # make sure the collection is created
        self.assertIn(
            'omegaml.dfgroup.datastore', store.mongodb.collection_names())
        # note column order can differ due to insertion order since pandas 0.25.1
        # hence using [] to ensure same column order for both expected, result
        df2 = store.get('dfgroup', kwargs={'b': 1})
        self.assertTrue(df2.equals(result_df[df2.columns]))
        df3 = store.get('dfgroup')
        self.assertTrue(df3.equals(df[df3.columns]))
        df4 = store.get('dfgroup', kwargs={'a': 1})
        self.assertTrue(df4.equals(result_df[df4.columns]))

    def test_store_dataframe_as_hdf(self):
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        df = pd.DataFrame(data)
        store = OmegaStore()
        meta = store.put(df, 'foo', as_hdf=True)
        self.assertEqual(meta.kind, 'pandas.hdf')
        # make sure the hdf file is actually there
        self.assertIn('omegaml.foo.hdf', store.fs.list())
        df2 = store.get('foo')
        self.assertTrue(df.equals(df2), "dataframes differ")
        # test for non-existent file raises exception
        meta = store.put(df2, 'foo_will_be_removed', as_hdf=True)
        file_id = store.fs.get_last_version(
            'omegaml.foo_will_be_removed.hdf')._id
        store.fs.delete(file_id)
        self.assertRaises(
            gridfs.errors.NoFile, store.get, 'foo_will_be_removed')
        store2 = OmegaStore()
        # test hdf file is not there
        self.assertNotIn('hdfdf.hdf', store2.fs.list())

    def test_put_same_name(self):
        """ test if metadata is updated instead of a new created """
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        df = pd.DataFrame(data)
        store = OmegaStore()
        # store the object
        meta = store.put(df, 'foo')
        # store it again
        meta2 = store.put(df, 'foo', append=False)
        # we should still have a new object in metadata
        # and the old should be gone
        self.assertNotEqual(meta.pk, meta2.pk)
        # Meta is to silence lint on import error
        Meta = store._Metadata
        metas = Meta.objects(name='foo', prefix=store.prefix,
                             bucket=store.bucket).all()
        self.assertEqual(len(metas), 1)
        df2 = store.get('foo')
        self.assertTrue(df.equals(df2))

    def test_put_append_false(self):
        """ test if we can create a new dataframe without previous metadata """
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        df = pd.DataFrame(data)
        store = OmegaStore()
        # store the object
        unique_name = uuid.uuid4().hex
        meta = store.put(df, unique_name, append=False)
        self.assertEqual(meta['name'], unique_name)

    def test_store_with_attributes(self):
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        df = pd.DataFrame(data)
        store = OmegaStore()
        # store the object, no attributes
        meta = store.put(df, 'foo', append=False)
        meta = store.metadata('foo')
        self.assertEqual(meta.attributes, {})
        # update attributes
        meta = store.put(df, 'foo', append=False, attributes={'foo': 'bar'})
        meta = store.metadata('foo')
        self.assertEqual(meta.attributes, {'foo': 'bar'})
        meta = store.put(
            df, 'foo', append=False, attributes={'foo': 'bax',
                                                 'foobar': 'barbar'})
        meta = store.metadata('foo')
        self.assertEqual(meta.attributes, {'foo': 'bax',
                                           'foobar': 'barbar'})

    def test_drop(self):
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        df = pd.DataFrame(data)
        store = OmegaStore()
        meta = store.put(df, 'hdfdf', as_hdf=True)
        self.assertTrue(store.drop('hdfdf'))
        meta = store.put(df, 'datadf')
        self.assertTrue(store.drop('datadf'))
        self.assertEqual(
            store.list('datadf'), [], 'expected the store to be empty')
        with self.assertRaises(DoesNotExist):
            store.drop('nxstore', force=False)
        try:
            store.drop('nxstore', force=True)
            raised = False
        except:
            raised = True
        self.assertFalse(raised)

    def test_list_raw(self):
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        df = pd.DataFrame(data)
        store = OmegaStore()
        meta = store.put(df, 'hdfdf', as_hdf=True)
        # list with pattern
        entries = store.list(pattern='hdf*', raw=True)
        self.assertTrue(isinstance(entries[0], store._Metadata))
        self.assertEqual('hdfdf', entries[0].name)
        self.assertEqual(len(entries), 1)
        # list with regexp
        entries = store.list(regexp='hdf.*', raw=True)
        self.assertTrue(isinstance(entries[0], store._Metadata))
        self.assertEqual('hdfdf', entries[0].name)
        self.assertEqual(len(entries), 1)
        # list without pattern nor regexp
        entries = store.list('hdfdf', kind=MDREGISTRY.PANDAS_HDF, raw=True)
        self.assertTrue(isinstance(entries[0], store._Metadata))
        self.assertEqual('hdfdf', entries[0].name)
        self.assertEqual(len(entries), 1)
        # subset kind
        entries = store.list('hdfdf', raw=True, kind=MDREGISTRY.PANDAS_DFROWS)
        self.assertEqual(len(entries), 0)
        entries = store.list('hdfdf', raw=True, kind=MDREGISTRY.PANDAS_HDF)
        self.assertEqual(len(entries), 1)

    def test_lazy_unique(self):
        """ test getting a MDataFrame and unique values """
        data = {
            'a': list(range(1, 10)),
            'b': list(range(1, 10))
        }
        df = pd.DataFrame(data)
        store = OmegaStore()
        meta = store.put(df, 'foo', append=False)
        val = store.get('foo', lazy=True).a.unique().value
        self.assertListEqual(data['a'], list(val))

    def test_store_series(self):
        """ test storing a pandas series with it's own index """
        from string import ascii_lowercase
        series = pd.Series(range(10), index=(c for c in ascii_lowercase[0:10]))
        store = OmegaStore()
        store.put(series, 'fooseries', append=False)
        series2 = store.get('fooseries')
        assert_series_equal(series, series2)

    def test_store_named_series(self):
        """ test storing a pandas series with it's own index """
        from string import ascii_lowercase
        series = pd.Series(range(10),
                           name='foo',
                           index=(c for c in ascii_lowercase[0:10]))
        store = OmegaStore()
        store.put(series, 'fooseries', append=False)
        series2 = store.get('fooseries')
        assert_series_equal(series, series2)

    def test_store_series_timeindex(self):
        """ test storing a pandas series with it's own index """
        series = pd.Series(range(10),
                           name='foo',
                           index=pd.date_range(pd.datetime(2016, 1, 1),
                                               pd.datetime(2016, 1, 10)))
        store = OmegaStore()
        store.put(series, 'fooseries', append=False)
        series2 = store.get('fooseries')
        assert_series_equal(series, series2)

    def test_store_irregular_column_names(self):
        """ test storing irregular column names """
        df = pd.DataFrame({'x_1': range(10)})
        store = OmegaStore()
        store.put(df, 'foo', append=False)
        df2 = store.get('foo')
        self.assertEqual(df.columns, df2.columns)

    def test_store_datetime(self):
        """ test storing naive datetimes """
        df = pd.DataFrame({
            'x': pd.date_range(pd.datetime(2016, 1, 1),
                               pd.datetime(2016, 1, 10))
        })
        store = OmegaStore()
        store.put(df, 'test-date', append=False)
        df2 = store.get('test-date')
        testing.assert_frame_equal(df, df2)

    def test_store_tz_datetime(self):
        """ test storing timezoned datetimes """
        df = pd.DataFrame({
            'y': pd.date_range('2019-10-01', periods=5, tz='US/Eastern', normalize=True)
        })
        store = OmegaStore()
        store.put(df, 'test-date', append=False)
        df2 = store.get('test-date')
        testing.assert_frame_equal(df, df2)

    # TODO support DST-crossing datetime objects. use UTC to avoid the issue
    @skip('date ranges across dst period start/end do not return the original DatetimeIndex values')
    def test_store_tz_datetime_dst(self):
        """ test storing timezoned datetimes """
        # 2019 11 03 02:00 is the end of US DST https://www.timeanddate.com/time/dst/2019.html
        # pymongo will transform the object into a naive dt at UTC time at +3h (arguably incorrectly so)
        # while pandas creates the Timestamp as UTC -4 (as the day starts at 00:00, not 02:00).
        # On rendering back to a tz-aware datetime, this yields the wrong date (1 day eaerlier) because
        # pandas applies -4 on converting from UTC to US/Eastern (correctly).
        df = pd.DataFrame({
            'y': pd.date_range('2019-11-01', periods=5, tz='US/Eastern', normalize=True)
        })
        store = OmegaStore()
        store.put(df, 'test-date', append=False)
        df2 = store.get('test-date')
        # currently this fails, see @skip reason
        testing.assert_frame_equal(df, df2)


    def test_store_dict_in_df(self):
        df = pd.DataFrame({
            'x': [{'foo': 'bar '}],
        })
        store = OmegaStore()
        store.put(df, 'test-dict', append=False)
        df2 = store.get('test-dict')
        testing.assert_frame_equal(df, df2)

    def test_existing_arbitrary_collection_flat(self):
        data = {'foo': 'bar',
                'bax': 'fox'}
        store = OmegaStore()
        store.register_backend(PandasRawDictBackend.KIND, PandasRawDictBackend)
        foo_coll = store.mongodb['foo']
        foo_coll.insert(data)
        store.make_metadata('myfoo', collection='foo', kind='pandas.rawdict').save()
        self.assertIn('myfoo', store.list())
        # test we get back _id column if raw=True
        data_df = store.get('myfoo', raw=True)
        data_raw = store.collection('myfoo').find_one()
        assert_frame_equal(json_normalize(data_raw), data_df)
        # test we get just the data column
        data_ = store.get('myfoo', raw=False)
        cols = ['foo', 'bax']
        assert_frame_equal(json_normalize(data)[cols], data_[cols])

    def test_existing_arbitrary_collection_nested(self):
        data = {'foo': 'bar',
                'bax': {
                    'fox': 'fax',
                }}
        store = OmegaStore()
        store.register_backend(PandasRawDictBackend.KIND, PandasRawDictBackend)
        foo_coll = store.mongodb['foo']
        foo_coll.insert(data)
        store.make_metadata('myfoo', collection='foo', kind='pandas.rawdict').save()
        self.assertIn('myfoo', store.list())
        # test we get back _id column if raw=True
        data_df = store.get('myfoo', raw=True)
        data_raw = store.collection('myfoo').find_one()
        assert_frame_equal(json_normalize(data_raw), data_df)
        # test we get just the data column
        data_ = store.get('myfoo', raw=False)
        cols = ['foo', 'bax.fox']
        assert_frame_equal(json_normalize(data)[cols], data_[cols])

    def test_existing_arbitrary_collection_mdataframe(self):
        data = {'foo': 'bar',
                'bax': {
                    'fox': 'fax',
                }}
        store = OmegaStore()
        store.register_backend(PandasRawDictBackend.KIND, PandasRawDictBackend)
        foo_coll = store.mongodb['foo']
        foo_coll.insert(data)
        store.make_metadata('myfoo', collection='foo', kind='pandas.rawdict').save()
        self.assertIn('myfoo', store.list())
        # test we get back _id column if raw=True
        mdf = store.getl('myfoo', raw=True)
        self.assertIsInstance(mdf, MDataFrame)
        data_df = mdf.value
        data_raw = store.collection('myfoo').find_one()
        assert_frame_equal(json_normalize(data_raw), data_df)
        # test we get just the data column
        mdf = store.getl('myfoo', raw=False)
        self.assertIsInstance(mdf, MDataFrame)
        data_df = mdf.value
        data_raw = store.collection('myfoo').find_one()
        cols = ['foo', 'bax.fox']
        assert_frame_equal(json_normalize(data)[cols], data_df[cols])

    def test_arbitrary_collection_new(self):
        data = {'foo': 'bar',
                'bax': 'fox'}
        store = OmegaStore()
        store.register_backend(PandasRawDictBackend.KIND, PandasRawDictBackend)
        # create the collection
        foo_coll = store.mongodb['foo']
        foo_coll.insert(data)
        # store the collection as is
        store.put(foo_coll, 'myfoo').save()
        self.assertIn('myfoo', store.list())
        # test we get back _id column if raw=True
        data_df = store.get('myfoo', raw=True)
        data_raw = store.collection('myfoo').find_one()
        assert_frame_equal(json_normalize(data_raw), data_df)
        # test we get just the data column
        data_df = store.get('myfoo', raw=False)
        data_raw = store.collection('myfoo').find_one()
        del data_raw['_id']
        assert_frame_equal(json_normalize(data_raw), data_df)
        cols = ['foo', 'bax']
        assert_frame_equal(data_df[cols], json_normalize(data_raw)[cols])

    def test_raw_files(self):
        store = OmegaStore()
        store.register_backend(PythonRawFileBackend.KIND, PythonRawFileBackend)
        # test we can write from a file-like object
        data = "some data"
        file_like = BytesIO(data.encode('utf-8'))
        store.put(file_like, 'myfile')
        self.assertEqual(data.encode('utf-8'), store.get('myfile').read())
        # test we can write from an actual file
        data = "some other data"
        file_like = BytesIO(data.encode('utf-8'))
        with open('/tmp/testfile.txt', 'wb') as fout:
            fout.write(file_like.read())
        store.put('/tmp/testfile.txt', 'myfile')
        self.assertEqual(data.encode('utf-8'), store.get('myfile').read())







import os

from omegaml.client.docoptparser import CommandBase
from omegaml.client.util import get_omega


class ScriptsCommandBase(CommandBase):
    """
    Usage:
        om scripts list [<pattern>] [options]
        om scripts put <path> <name> [options]
        om scripts delete <name> [options]
    """
    command = 'scripts'

    def list(self):
        om = get_omega(self.args)
        raw = self.args.get('--raw', False)
        self.logger.info(om.scripts.list(raw=raw))

    def put(self):
        om = get_omega(self.args)
        script_path = self.args.get('<path>')
        name = self.args.get('<name>')
        if os.path.exists(script_path):
            name = name or os.path.basename(script_path)
            abs_path = os.path.abspath(script_path)
            meta = om.scripts.put('pkg://{}'.format(abs_path), name)
            self.logger.info(meta)
        else:
            raise ValueError('{} is not a valid path'.format(script_path))

    def delete(self):
        om = get_omega(self.args)
        name = self.args.get('<name>')
        om.scripts.drop(name, force=True)




def hello(**kwargs):
    return "hello from helloworld", kwargs


def run(om, **kwargs):
    """
    the script API execution entry point
    :return: result
    """
    result = hello(**kwargs)
    return result


from itertools import product

import hashlib
import json
from uuid import uuid4

import pandas as pd
import six
from omegaml import defaults
from omegaml.documents import make_QueryCache
from omegaml.mdataframe import MDataFrame, MSeries
from omegaml.store import qops
from omegaml.util import make_tuple, extend_instance


class ApplyMixin(object):
    """
    Implements the apply() mixin supporting arbitrary functions to build aggregation pipelines

    Note that .apply() does not execute immediately. Instead it builds an aggregation pipeline
    that is executed on MDataFrame.value. Note that .apply() calls cannot be cascaded yet, i.e.
    a later .apply() will override a previous.apply().

    See ApplyContext for usage examples.
    """

    def __init__(self, *args, **kwargs):
        super(ApplyMixin, self).__init__(*args, **kwargs)
        self._init_mixin(*args, **kwargs)

    def _init_mixin(self, *args, **kwargs):
        self.apply_fn = kwargs.get('apply_fn', None)
        # set to True if the pipeline is a facet operation
        self.is_from_facet = kwargs.get('is_from_facet', False)
        # index columns
        self.index_columns = kwargs.get('index_columns', [])
        # cache used on persist()
        self.cache = kwargs.get('cache', ApplyCache())

    def nocache(self):
        self.cache = None
        return self

    def reset_cache(self, full=False):
        """
        Reset the apply cache

        :param full: if True will reset all caches for the collection, if False will only remove
            the cache for the specific .apply operations
        :return:
        """
        QueryCache = make_QueryCache()
        if full:
            QueryCache.objects.filter(value__collection=self.collection.name).delete()
        else:
            pipeline = self._build_pipeline()
            key = self._make_cache_key(self.collection, pipeline)
            QueryCache.objects.filter(key=key).delete()
        return self

    def _make_cache_key(self, collection, pipeline):
        # remove random output value
        if '$out' in pipeline[-1] and pipeline[-1]['$out'].startswith('cache'):
            pipeline = list(pipeline)[:-1]
        spipeline = json.dumps(pipeline, sort_keys=True)
        data = '{}_{}'.format(collection.name, spipeline).encode('utf-8')
        key = hashlib.md5(data).hexdigest()
        return key

    def _getcopy_kwargs(self, **kwargs):
        kwargs = super(ApplyMixin, self)._getcopy_kwargs(**kwargs)
        kwargs.update(is_from_facet=self.is_from_facet,
                      index_columns=self.index_columns,
                      cache=self.cache,
                      apply_fn=self.apply_fn)
        return kwargs

    def noapply(self):
        self.apply_fn = None
        return self

    def apply(self, fn, inplace=False, preparefn=None):
        if inplace:
            obj = self
        else:
            kwargs = self._getcopy_kwargs()
            kwargs.update(preparefn=preparefn)
            if isinstance(self, MSeries):
                obj = MSeries(self.collection, **kwargs)
            else:
                obj = MDataFrame(self.collection, **kwargs)
        obj.apply_fn = fn
        return obj

    def persist(self):
        """
        Execute and store results in cache

        Any pipeline of the same operations, in the same order, on
        the same collection will return the same result.
        """
        # generate a cache key
        pipeline = self._build_pipeline()
        key = self._make_cache_key(self.collection, pipeline)
        outname = 'cache_{}'.format(uuid4().hex)
        value = {
            'collection': self.collection.name,
            'result': outname,
        }
        # do usual processing, store result
        # -- note we pass pipeline to avoid processing iterators twice
        pipeline.append({
            '$out': outname,
        })
        cursor = self._get_cursor(pipeline=pipeline, use_cache=False)
        # consume cursor to store output (via $out)
        for v in cursor:
            pass
        # set cache
        self.cache.set(key, value)
        return key

    def set_index(self, columns):
        self.index_columns = make_tuple(columns)
        return self

    def inspect(self, explain=False, *args, **kwargs):
        if self.apply_fn:
            details = {
                'pipeline': self._build_pipeline()
            }
            if explain:
                details.update(self.__dict__)
            return details
        return super(ApplyMixin, self).inspect(*args, explain=explain, **kwargs)

    def _execute(self):
        ctx = ApplyContext(self, columns=self.columns)
        try:
            result = self.apply_fn(ctx)
        except Exception as e:
            msg = [repr(stage) for stage in ctx.stages] + [repr(e)]
            raise RuntimeError(msg)
        if result is None or isinstance(result, ApplyContext):
            result = result or ctx
            self.index_columns = self.index_columns or result.index_columns
            return result
        elif isinstance(result, list):
            return result
        elif isinstance(result, dict):
            # expect a mapping of col=ApplyContext each with its own list of stages
            # -- build a combined context by adding each expression
            #    this ensures any multi-stage projections are carried forward
            facets = {}
            for col, expr in six.iteritems(result):
                if isinstance(expr, ApplyContext):
                    facets[col] = list(expr)
                    project = {
                        '$project': {
                            col: '$' + expr.columns[0]
                        },
                    }
                    facets[col].append(project)
                else:
                    facets[col] = expr
            facet = {
                '$facet': facets
            }
            self.is_from_facet = True
            return [facet]
        raise ValueError('Cannot build pipeline from apply result of type {}'.format(type(result)))

    def _build_pipeline(self):
        pipeline = []
        stages = self._execute()
        pipeline.extend(stages)
        self._amend_pipeline(pipeline)
        return pipeline

    def _amend_pipeline(self, pipeline):
        """ amend pipeline with default ops on coll.aggregate() calls """
        if self.sort_order:
            sort = qops.SORT(**dict(qops.make_sortkey(self.sort_order)))
            pipeline.append(sort)
        return pipeline

    def _get_cached_cursor(self, pipeline=None, use_cache=True):
        pipeline = pipeline or self._build_pipeline()
        if use_cache and self.cache:
            key = self._make_cache_key(self.collection, pipeline)
            entry = self.cache.get(key)
            if entry is not None:
                # read result
                outname = entry.value['result']
                return self.collection.database[outname].find()

    def _get_cursor(self, pipeline=None, use_cache=True):
        # for apply functions, call the apply function, expecting a pipeline in return
        if self.apply_fn:
            pipeline = pipeline or self._build_pipeline()
            cursor = self._get_cached_cursor(pipeline=pipeline, use_cache=use_cache)
            if cursor is None:
                filter_criteria = self._get_filter_criteria()
                cursor = self.collection.aggregate(pipeline, filter=filter_criteria, allowDiskUse=True)
        else:
            cursor = super(ApplyMixin, self)._get_cursor()
        return cursor

    def _get_dataframe_from_cursor(self, cursor):
        df = super(ApplyMixin, self)._get_dataframe_from_cursor(cursor)
        if self.is_from_facet:
            # if this was from a facet pipeline (i.e. multi-column mapping), combine
            # $facet returns one document for each stage.
            frames = []
            for col in df.columns:
                coldf = pd.DataFrame(df[col].iloc[0]).set_index('_id')
                frames.append(coldf)
            df = pd.concat(frames, axis=1).reset_index()
            df = self._restore_dataframe_proper(df)
        # TODO write a unit test for this condition
        if self.index_columns and all(col in df.columns for col in self.index_columns):
            df.set_index(self.index_columns, inplace=True)
        return df


class ApplyContext(object):
    """
    Enable apply functions

    .apply(fn) will call fn(ctx) where ctx is an ApplyContext.
    The context supports methods to apply functions in a Pandas-style apply manner. ApplyContext is extensible
    by adding an extension class to defaults.OMEGA_MDF_APPLY_MIXINS.

    Note that unlike a Pandas DataFrame, ApplyContext does not itself contain any data.
    Rather it is part of an expression tree, i.e. the aggregation pipeline. Thus any
    expressions applied are translated into operations on the expression tree. The expression
    tree is evaluated on MDataFrame.value, at which point the ApplyContext nor the function
    that created it are active.

    Examples::

        mdf.apply(lambda v: v * 5 ) => multiply every column in dataframe
        mdf.apply(lambda v: v['foo'].dt.week) => get week of date for column foo
        mdf.apply(lambda v: dict(a=v['foo'].dt.week,
                                 b=v['bar'] * 5) => run multiple pipelines and get results

        The callable passed to apply can be any function. It can either return None,
        the context passed in or a list of pipeline stages.

        # apply any of the below functions
        mdf.apply(customfn)

        # same as lambda v: v.dt.week
        def customfn(ctx):
            return ctx.dt.week

        # simple pipeline
        def customfn(ctx):
            ctx.project(x={'$multiply: ['$x', 5]})
            ctx.project(y={'$divide: ['$x', 2]})

        # complex pipeline
        def customfn(ctx):
            return [
                { '$match': ... },
                { '$project': ... },
            ]
    """

    def __init__(self, caller, columns=None, index=None):
        self.caller = caller
        self.columns = columns
        self.index_columns = index or []
        self.computed = []
        self.stages = []
        self.expressions = []
        self._apply_mixins()

    def _apply_mixins(self):
        """
        apply mixins in defaults.OMEGA_MDF_APPLY_MIXINS
        """
        from omegaml import settings
        defaults = settings()
        for mixin, applyto in defaults.OMEGA_MDF_APPLY_MIXINS:
            if any(v in self.caller._applyto for v in applyto.split(',')):
                extend_instance(self, mixin)

    def __iter__(self):
        # return pipeline stages
        for stage in self.stages:
            if isinstance(stage, ApplyContext):
                for sub_stage in stage:
                    yield sub_stage
            else:
                yield stage

    def __getitem__(self, sel):
        """
        return a stage subset on a column
        """
        subctx = ApplyContext(self.caller, columns=make_tuple(sel), index=self.index_columns)
        self.add(subctx)
        return subctx

    def __setitem__(self, sel, val):
        """
        add a projection to a sub context

        ctx['col'] = value-expression
        """
        mapping = {
            col: v
            for (col, v) in zip(make_tuple(sel), make_tuple(val))}
        self.project(mapping)

    def __repr__(self):
        return 'ApplyContext(stages={}, expressions={})'.format(self.stages, self.expressions)

    def add(self, stage):
        """
        Add a processing stage to the pipeline

        see https://docs.mongodb.com/manual/meta/aggregation-quick-reference/
        """
        self.stages.append(stage)
        return self

    def project_keeper_columns(self):
        # keep index, computed
        index = {
            col: '$' + col
            for col in self.index_columns}
        computed = {
            col: '$' + col
            for col in self.computed}
        keep = {}
        keep.update(index)
        keep.update(computed)
        project = self.project(keep, keep=True)
        return project

    def _getLastStageKind(self, kind):
        # see if there is already an open projection stage
        for stage in self.stages[::-1]:
            if kind in stage:
                return stage

    def _getProjection(self, append=False):
        stage = self._getLastStageKind('$project')
        if stage is None or append:
            stage = {
                '$project': {
                    '_id': 1,
                }
            }
            self.stages.append(stage)
        return stage

    def _getGroupBy(self, by=None, append=False):
        stage = self._getLastStageKind('$group')
        if stage and stage['$group']['_id'] != by and by != '$$last':
            # if a different groupby criteria, add a new one
            stage = None
        if stage is None and by == '$$last':
            by = None
        if stage is None or append:
            stage = {
                '$group': {
                    '_id': by,
                }
            }
            self.stages.append(stage)
        return stage

    def groupby(self, by, expr=None, append=None, **kwargs):
        """
        add a groupby accumulation using $group

        :param by: the groupby columns, if provided as a list will be transformed
        :param expr:
        :param append:
        :param kwargs:
        :return:

        """
        by = make_tuple(by)
        self.index_columns = self.index_columns + list(by)
        # define groupby
        by = {col: '$' + col for col in by}
        stage = self._getGroupBy(by)
        groupby = stage['$group']
        # add acccumulators
        expr = expr or {
            col: colExpr
            for col, colExpr in six.iteritems(kwargs)}
        groupby.update(expr)
        # add a projection to extract groupby values
        extractId = {
            col: '$_id.' + col
            for col in by}
        # add a projection to keep accumulator columns
        keepCols = {
            col: 1
            for col in expr}
        keepCols.update(extractId)
        self.project(keepCols, append=True)
        # sort by groupby keys
        self.add({
            '$sort': {
                col: 1
                for col in by}
        })
        return self

    def project(self, expr=None, append=False, keep=False, **kwargs):
        """
        add a projection using $project

        :param expr: the column-operator mapping
        :param append: if True add a $project stage, otherwise add to existing
        :param kwargs: if expr is None, the column-operator mapping as kwargs
        :return: ApplyContext

        """
        # get last $project stage in pipeline
        stage = self._getProjection(append=append)
        expr = expr or kwargs
        self.expressions.append(expr)
        for k, v in six.iteritems(expr):
            # only append to stage if no other column projection was there
            project = stage.get('$project')
            if k not in project:
                project.update({
                    k: v
                })
            elif not keep:
                # if a column is already projected, add a new projection stage
                stage = self._getProjection(append=True)
                project = stage.get('$project')
                project.update({
                    k: v
                })
        return self


class ApplyArithmetics(object):
    """
    Math operators for ApplyContext

    * :code:`__mul__` (*)
    * :code:`__add__` (+)
    * :code:`__sub__` (-)
    * :code:`__div__` (/)
    * :code:`__floordiv__` (//)
    * :code:`__mod__` (%)
    * :code:`__pow__` (pow)
    * :code:`__ceil__` (ceil)
    * :code:`__floor__` (floor)
    * :code:`__trunc__` (trunc)
    * :code:`__abs__` (abs)
    * :code:`sqrt` (math.sqrt)

    """

    def __arithmop__(op, wrap_op=None):
        """
        return a pipeline $project stage math operator as
           { col:
              { '$operator': [ values, ...] }
              ...
           }

        If wrap_op is specified, will wrap the $operator clause as
           { col:
              { '$wrap_op': { '$operator': [ values, ...] } }0
              ...
           }
        """

        def inner(self, other):
            terms = []
            for term in make_tuple(other):
                if isinstance(term, six.string_types):
                    term = '$' + term
                terms.append(term)
            def wrap(expr):
                if wrap_op is not None:
                    expr = {
                        wrap_op: expr
                    }
                return expr
            mapping = {
                col: wrap({
                    op: ['$' + col] + terms,
                }) for col in self.columns}
            keepCols = {
                col: '$' + col
                for col in self.index_columns}
            mapping.update(keepCols)
            self.project(mapping)
            return self

        return inner

    #: multiply
    __mul__ = __arithmop__('$multiply')
    #: add
    __add__ = __arithmop__('$add')
    #: subtract
    __sub__ = __arithmop__('$subtract')
    #: divide
    __div__ = __arithmop__('$divide')
    __truediv__ = __arithmop__('$divide')
    #: divide integer
    __floordiv__ = __arithmop__('$divide', wrap_op='$floor')
    #: modulo (%)
    __mod__ = __arithmop__('$mod')
    #: pow
    __pow_ = __arithmop__('$pow')
    #: ceil
    __ceil__ = __arithmop__('$ceil')
    #: floor
    __floor__ = __arithmop__('$floor')
    #: truncate
    __trunc__ = __arithmop__('$trunc')
    #: absolute
    __abs__ = __arithmop__('$abs')
    #: square root
    sqrt = __arithmop__('sqrt')


class ApplyDateTime(object):
    """
    Datetime operators for ApplyContext
    """

    @property
    def dt(self):
        return self

    def __dtop__(op):
        """
        return a datetime $project operator as
           { col:
              { '$operator': '$col} }
              ...
           }
        """

        def inner(self, columns=None):
            columns = make_tuple(columns or self.columns)
            mapping = {
                col: {
                    op: '$' + col,
                }
                for col in columns}
            self.project(mapping)
            return self

        inner.__doc__ = op.replace('$', '')
        return inner

    # mongodb mappings
    _year = __dtop__('$year')
    _month = __dtop__('$month')
    _week = __dtop__('$week')
    _dayOfWeek = __dtop__('$dayOfWeek')
    _dayOfMonth = __dtop__('$dayOfMonth')
    _dayOfYear = __dtop__('$dayOfYear')
    _hour = __dtop__('$hour')
    _minute = __dtop__('$minute')
    _second = __dtop__('$second')
    _millisecond = __dtop__('$millisecond')
    _isoDayOfWeek = __dtop__('$isoDayOfWeek')
    _isoWeek = __dtop__('$isoWeek')
    _isoWeekYear = __dtop__('$isoWeekYear')

    # .dt accessor convenience similar to pandas.dt
    # see https://pandas.pydata.org/pandas-docs/stable/api.html#datetimelike-properties
    year = property(_year)
    month = property(_month)
    day = property(_dayOfMonth)
    hour = property(_hour)
    minute = property(_minute)
    second = property(_second)
    millisecond = property(_millisecond)
    week = property(_isoWeek)
    dayofyear = property(_dayOfYear)
    dayofweek = property(_dayOfWeek)


class ApplyString(object):
    """
    String operators
    """

    @property
    def str(self):
        return self

    def __strexpr__(op, unwind=False, base=None, max_terms=None):
        """
        return a pipeline $project string operator as
           { col:
              { '$operator': [ values, ...] }
              ...
           }
        """

        def inner(self, other, *args):
            # get all values passed and build terms from them
            values = list(make_tuple(other) + args)
            terms = []
            for term in values:
                if isinstance(term, six.string_types):
                    # if the term is a column name, add as a column name
                    if term in self.columns:
                        term = '$' + term
                    # allow to specify values explicitely by $$<value> => <value>
                    term = term.replace('$$', '')
                terms.append(term)
            # limit number of terms if requested
            if max_terms:
                terms = terms[:max_terms]
            # add projection of output columns to operator
            mapping = {
                col: {
                    op: terms if base is None else ['$' + col] + terms,
                } for col in self.columns}
            self.project(mapping)
            # unwind all columns if requested
            if unwind:
                exprs = [{'$unwind': {
                    'path': '$' + col
                }} for col in self.columns]
                self.stages.extend(exprs)
            return self

        inner.__doc__ = op.replace('$', '')
        return inner

    def __strunary__(op, unwind=False):
        """
        return a datetime $project operator as
           { col:
              { '$operator': '$col} }
              ...
           }
        """

        def inner(self, columns=None):
            columns = make_tuple(columns or self.columns)
            mapping = {
                col: {
                    op: '$' + col,
                }
                for col in columns}
            self.project(mapping)
            if unwind:
                self.stages.append({
                    '$unwind': {
                        ''
                    }
                })
            return self

            inner.__doc__ = op.replace('$', '')

        return inner

    def isequal(self, other):
        self.strcasecmp(other)
        # strcasecmp returns 0 for equality, 1 and -1 for greater/less than
        # https://docs.mongodb.com/manual/reference/operator/aggregation/strcasecmp/
        mapping = {
            col: {
                '$cond': {
                    'if': {'$eq': ['$' + col, 0]},
                    'then': True,
                    'else': False,
                }
            }
            for col in self.columns}
        self.project(mapping)

    concat = __strexpr__('$concat', base=True)
    split = __strexpr__('$split', unwind=True, base=True, max_terms=2)
    upper = __strunary__('$toUpper')
    lower = __strunary__('$toLower')
    substr = __strexpr__('$substr', base=True)
    strcasecmp = __strexpr__('$strcasecmp', base=True)
    len = __strunary__('$strLenBytes')
    index = __strexpr__('$indexOfBytes', base=True)


class ApplyAccumulators(object):
    def agg(self, map=None, **kwargs):
        stage = self._getGroupBy(by='$$last')
        specs = map or kwargs
        for col, colExpr in six.iteritems(specs):
            if isinstance(colExpr, dict):
                # specify an arbitrary expression
                groupby = stage['$group']
                groupby[col] = colExpr
            elif isinstance(colExpr, six.string_types):
                # specify some known operator
                if hasattr(self, colExpr):
                    method = getattr(self, colExpr)
                    method(col)
                else:
                    raise SyntaxError('{} is not known'.format(colExpr))
            elif isinstance(colExpr, (tuple, list)):
                # specify a list of some known operators
                for statExpr in colExpr:
                    if hasattr(self, statExpr):
                        method = getattr(self, statExpr)
                        method(col)
                    else:
                        raise SyntaxError('{} is not known'.format(statExpr))
            elif callable(colExpr):
                # specify a callable that returns an expression
                groupby = stage['$group']
                groupby[col] = colExpr(col)
            else:
                SyntaxError('{} on column {} is unknown or invalid'.format(colExpr, col))
        return self

    def __statop__(op, opname=None):
        opname = opname or op.replace('$', '')

        def inner(self, columns=None):
            columns = make_tuple(columns or self.columns)
            stage = self._getGroupBy(by='$$last')
            groupby = stage['$group']
            groupby.update({
                               '{}_{}'.format(col, opname): {
                                   op: '$' + col
                               } for col in columns
                               })
            self.computed.extend(groupby.keys())
            self.project_keeper_columns()
            return self

        return inner

    sum = __statop__('$sum')
    avg = __statop__('$avg')
    mean = __statop__('$avg')
    min = __statop__('$min')
    max = __statop__('$max')
    std = __statop__('$stdDevSamp', 'std')


class ApplyCache(object):
    """
    A Cache that works on collections and pipelines
    """

    def set(self, key, value):
        # https://stackoverflow.com/a/22003440/890242
        QueryCache = make_QueryCache()
        QueryCache.objects(key=key).update_one(set__key="{}".format(key),
                                               set__value=value, upsert=True)

    def get(self, key):
        QueryCache = make_QueryCache()
        try:
            result = QueryCache.objects.get(key=key)
        except:
            result = None
        return result


class ApplyStatistics(object):
    def quantile(self, q=.5):
        def preparefn(val):
            return val.pivot('percentile', 'var', 'value')
        return self.apply(self._percentile(q), preparefn=preparefn)

    def cov(self):
        def preparefn(val):
            val = val.pivot('x', 'y', 'cov')
            val.index.name = None
            val.columns.name = None
            return val
        return self.apply(self._covariance, preparefn=preparefn)

    def corr(self):
        def preparefn(val):
            val = val.pivot('x', 'y', 'rho')
            val.index.name = None
            val.columns.name = None
            return val
        return self.apply(self._pearson, preparefn=preparefn)

    def _covariance(self, ctx):
        # this works
        # source http://ci.columbia.edu/ci/premba_test/c0331/s7/s7_5.html
        facets = {}
        means = {}
        unwinds = []
        count = len(ctx.caller.noapply()) - 1
        for x, y in product(ctx.columns, ctx.columns):
            xcol = '$' + x
            ycol = '$' + y
            # only calculate the same column's mean once
            if xcol not in means:
                means[xcol] = ctx.caller[x].noapply().mean().values[0, 0]
            if ycol not in means:
                means[ycol] = ctx.caller[y].noapply().mean().values[0, 0]
            sumands = {
                xcol: {
                    '$subtract': [xcol, means[xcol]]
                },
                ycol: {
                    '$subtract': [ycol, means[ycol]]
                }
            }
            multiply = {
                '$multiply': [sumands[xcol], sumands[ycol]]
            }
            agg = {
                '$group': {
                    '_id': None,
                    'value': {
                        '$sum': multiply
                    }
                }
            }
            project = {
                '$project': {
                    'cov': {
                        '$divide': ['$value', count],
                    },
                    'x': x,
                    'y': y,
                }
            }
            pipeline = [agg, project]
            outcol = '{}_{}'.format(x, y)
            facets[outcol] = pipeline
            unwinds.append({'$unwind': '$' + outcol})
        facet = {
            '$facet': facets,
        }
        expand = [{
            '$project': {
                'value': {
                    '$objectToArray': '$$CURRENT',
                }
            }
        }, {
            '$unwind': '$value'
        }, {
            '$replaceRoot': {
                'newRoot': '$value.v'
            }
        }]
        return [facet, *unwinds, *expand]

    def _pearson(self, ctx):
        # this works
        # source http://ilearnasigoalong.blogspot.ch/2017/10/calculating-correlation-inside-mongodb.html
        facets = {}
        unwinds = []
        for x, y in product(ctx.columns, ctx.columns):
            xcol = '$' + x
            ycol = '$' + y
            sumcolumns = {'$group': {'_id': None,
                                     'count': {'$sum': 1},
                                     'sumx': {'$sum': xcol},
                                     'sumy': {'$sum': ycol},
                                     'sumxsquared': {'$sum': {'$multiply': [xcol, xcol]}},
                                     'sumysquared': {'$sum': {'$multiply': [ycol, ycol]}},
                                     'sumxy': {'$sum': {'$multiply': [xcol, ycol]}}
                                     }}

            multiply_sumx_sumy = {'$multiply': ["$sumx", "$sumy"]}
            multiply_sumxy_count = {'$multiply': ["$sumxy", "$count"]}
            partone = {'$subtract': [multiply_sumxy_count, multiply_sumx_sumy]}

            multiply_sumxsquared_count = {'$multiply': ["$sumxsquared", "$count"]}
            sumx_squared = {'$multiply': ["$sumx", "$sumx"]}
            subparttwo = {'$subtract': [multiply_sumxsquared_count, sumx_squared]}

            multiply_sumysquared_count = {'$multiply': ["$sumysquared", "$count"]}
            sumy_squared = {'$multiply': ["$sumy", "$sumy"]}
            subpartthree = {'$subtract': [multiply_sumysquared_count, sumy_squared]}

            parttwo = {'$sqrt': {'$multiply': [subparttwo, subpartthree]}}

            rho = {'$project': {
                'rho': {
                    '$divide': [partone, parttwo]
                },
                'x': x,
                'y': y
            }}
            pipeline = [sumcolumns, rho]
            outcol = '{}_{}'.format(x, y)
            facets[outcol] = pipeline
            unwinds.append({'$unwind': '$' + outcol})
        facet = {
            '$facet': facets,
        }
        expand = [{
            '$project': {
                'value': {
                    '$objectToArray': '$$CURRENT',
                }
            }
        }, {
            '$unwind': '$value'
        }, {
            '$replaceRoot': {
                'newRoot': '$value.v'
            }
        }]
        return [facet, *unwinds, *expand]

    def _percentile(self, pctls=None):
        """
        calculate percentiles for all columns
        """
        pctls = pctls or [.25, .5, .75]
        if not isinstance(pctls, (list, tuple)):
                pctls = [pctls]

        def calc(col, p, outcol):
            # sort values
            sort = {
                '$sort': {
                    col: 1,
                }
            }
            # group/push to get an array of all values
            group = {
                '$group': {
                    '_id': col,
                    'values': {
                        '$push': "$" + col
                    },
                }
            }
            # find value at requested percentile
            perc = {
                '$arrayElemAt': [
                    '$values', {
                        '$floor': {
                        '$multiply': [{
                            '$size': '$values'
                        }, p]
                    }}
                ]
            }
            # map percentile value to output column
            project = {
                '$project': {
                    'var': col,
                    'percentile': 'p{}'.format(p),
                    'value': perc,
                }
            }
            return [sort, group, project]

        def inner(ctx):
            # for each column and requested percentile, build a pipeline
            # all pipelines will be combined into a $facet stage to
            # calculate every column/percentile tuple in parallel
            facets = {}
            unwind = []
            # for each column build a pipeline to calculate the percentiles
            for col in ctx.columns:
                for p in pctls:
                    # e.g. outcol for perc .25 of column abc => abcp25
                    outcol = '{}_p{}'.format(col, p).replace('0.', '')
                    facets[outcol] = calc(col, p, outcol)
                    unwind.append({'$unwind': '$'+ outcol})
            # process per-column pipelines in parallel, resulting in one
            # document for each variable + percentile combination
            facet = {
                '$facet': facets
            }
            # expand single document into one document per variable + percentile combo
            # the resulting set of documents contains var/percentile/value
            expand = [{
                '$project': {
                    'value': {
                        '$objectToArray': '$$CURRENT',
                    }
                }
            }, {
                '$unwind': '$value'
            }, {
                '$replaceRoot': {
                    'newRoot': '$value.v'
                }
            }]
            pipeline = [facet, *unwind, *expand]
            return pipeline

        return inner





import os

from omegaml.backends.keras import KerasBackend
from omegaml.util import temp_filename


class TensorflowKerasBackend(KerasBackend):
    KIND = 'tfkeras.h5'

    @classmethod
    def supports(self, obj, name, **kwargs):
        import tensorflow as tf
        tfSequential = tf.keras.models.Sequential
        tfModel = tf.keras.models.Model
        return isinstance(obj, (tfSequential, tfModel)) and not kwargs.get('as_savedmodel')

    def _save_model(self, model, fn):
        # override to implement model saving
        import tensorflow as tf
        from tensorflow import keras
        if tf.executing_eagerly():
            self._fix_model_for_saving(model)
        keras.models.save_model(model, fn)

    def _fix_model_for_saving(self, model):
        # see
        import tensorflow as tf
        from tensorflow.python.keras import backend as K
        with K.name_scope(model.optimizer.__class__.__name__):
            try:
                for i, var in enumerate(model.optimizer.weights):
                    name = 'variable{}'.format(i)
                    model.optimizer.weights[i] = tf.Variable(var, name=name)
            except NotImplementedError:
                pass

    def _load_model(self, fn):
        # override to implement model loading
        from tensorflow import keras
        load_model = keras.engine.saving.load_model
        return load_model(fn)

    def _load_model(self, fn):
        # override to implement model loading
        from tensorflow import keras
        return keras.models.load_model(fn)

    def fit(self, modelname, Xname, Yname=None, pure_python=True, tpu_specs=None, **kwargs):
        meta = self.model_store.metadata(modelname)
        tpu_specs = tpu_specs or meta.attributes.get('tpu_specs')
        if tpu_specs:
            try:
                result = self._fit_tpu(modelname, Xname, Yname, tpu_specs=tpu_specs, **kwargs)
            except:
                import logging
                logger = logging.getLogger(__name__)
                logger.warning('Error in _fit_tpu, reverting to fit on CPU')
            else:
                return result
        result = super(TensorflowKerasBackend, self).fit(modelname, Xname, Yname=Yname, pure_python=pure_python,
                                                         **kwargs)
        return result

    def _fit_tpu(self, modelname, Xname, Yname=None, tpu_specs=None, **kwargs):
        import tensorflow as tf
        # adopted from https://www.dlology.com/blog/how-to-train-keras-model-x20-times-faster-with-tpu-for-free/
        # This address identifies the TPU we'll use when configuring TensorFlow.
        # FIXME this will fail in tf 2.0, see https://github.com/tensorflow/tensorflow/issues/24412#issuecomment-491980177
        assert tf.__version__.startswith('1.'), "TPU only supported on tf < 2.0"
        tpu_device = tpu_specs or os.environ.get('COLAB_TPU_ADDR', '')
        assert tpu_device, "there is no TPU device"
        if tpu_device.startswith('grpc://'):
            tpu_worker = tpu_device
        else:
            tpu_worker = 'grpc://' + tpu_device
        tf.logging.set_verbosity(tf.logging.INFO)
        model = self.get_model(modelname)
        tpu_model = tf.contrib.tpu.keras_to_tpu_model(
            model,
            strategy=tf.contrib.tpu.TPUDistributionStrategy(
                tf.contrib.cluster_resolver.TPUClusterResolver(tpu_worker)))
        X = self.data_store.get(Xname)
        Y = self.data_store.get(Yname)
        tpu_model.fit(X, Y)
        fn = temp_filename()
        tpu_model.save_weights(fn, overwrite=True)
        model.load_weights(fn)
        meta = self.put_model(model, modelname)
        return meta



import os
import string
import sys
from collections import defaultdict

import numpy as np
import pandas as pd
from matplotlib.pyplot import imsave
from nbformat import v4
from sklearn.linear_model import LinearRegression

from omegaml.client import cli


class CliTestScenarios:
    """TestCase plugin with common environment scenarios"""

    def setUp(self):
        super().setUp()
        self.pretend_logger = None

    def get_test_logger(self):
        """
        create a test logger that captures info, error, debug output

        Returns:
            a logger object, use logger.data['info|error|debug'] to get
            respective logger output by cli
        """

        class TestLogger:
            data = defaultdict(list)

            def info(self, *args):
                self.data['info'].append(args)

            def error(self, *args):
                self.data['error'].append(args)

            def debug(self, *args):
                self.data['debug'].append(args)

            def clear(self):
                self.data = defaultdict(list)

        return TestLogger()

    def pretend_log(self, *args, start_empty=False, level='info'):
        """
        return what would be written in log

        Use to generate expected test output feasible for assertLogEqual
        This way you don't need to understand the internal of TestLogger

        Args:
            args: what would be passed to logger.info(args)

        Usage:
            you would like to test if the cli writes a log entry of some kind:

            # you expect the cli to issue a log statement like this
            logger.info('there were 0 items in the list')

            # code to write
            expected = self.pretend_log('there were 0 items in the list')
            self.assertLogEqual('info', expected)

        Returns:
            the actual log entry that would be written
        """
        if self.pretend_logger is None:
            self.pretend_logger = self.get_test_logger()
        logger = self.pretend_logger
        if start_empty:
            logger.clear()
        level_logger = getattr(logger, level)
        level_logger(*args)
        return logger.data[level]

    def get_log(self, level):
        return self.cli_logger.data[level]

    def assertLogSize(self, level, size):
        """ assert log of level has number of entries (lines) """
        self.assertEqual(size, len(self.get_log(level)))

    def assertLogEqual(self, level, expected):
        """ assert log of level has expected contents """
        self.assertEqual(expected, self.get_log(level))

    def assertLogContains(self, level, expected, compare_as=str):
        """
        assert the log level contains some content

        Each log entry is compared to expected, if no match is found
        will raise AssertionError. By default the comparision is
        a test of str(expected) in str(log) for every entry in expected
        and log respectively. To compare as object, pass compare_as=object

        Args:
            level: the log level
            expected: the expected entry(ies), as returned by pretend_log(). if
               a string is passed will be replaced with the result of
               pretend_log(level, expected)
            compare_as: type or callable for casting entries, defaults to str

        Raises:
            AssertionError if no match is found
        """
        if isinstance(expected, str):
            expected = self.pretend_log(expected, level=level)[-1:]
        data = self.get_log(level)
        if compare_as is not object:
            # unpack log entries, not a log entry is a *args tuple
            # hence we need to iterate that to get each value
            expected = [compare_as(v) for args in expected for v in args]
            data = [compare_as(v) for args in data for v in args]
        match = False
        for e in expected:
            for d in data:
                match |= e in d
        if not match:
            raise AssertionError(f'{expected} is not in {data} compared as {compare_as}')

    def cli(self, argv, new_start=False, **kwargs):
        self.cli_logger = self.get_test_logger()
        if new_start:
            self.cli_logger.clear()
            self.pretend_logger.clear() if self.pretend_logger else None
        if isinstance(argv, str):
            argv = argv.split(' ')
        self.cli_parser = cli.main(argv=argv, logger=self.cli_logger, **kwargs)

    def make_model(self, name):
        """
        create and store a model

        Args:
            name: the model name

        Returns:
            Metadata of the model
        """
        om = self.om
        # add a model, see that we get it back
        reg = LinearRegression()
        return om.models.put(reg, 'reg')

    def make_dataset_from_dataframe(self, name, N=100):
        """
        create and store a pandas dataframe

        Args:
            name: the dataset name

        Returns:
            Metadata of the dataset
        """
        import pandas as pd
        df = pd.DataFrame({
            'x': range(N),
            'y': range(N)
        })
        df['y'] = df['x'] * 2
        return self.om.datasets.put(df, name, append=False)

    def create_local_csv(self, path, size=(100, 2), sep=','):
        """
        create a local dataset

        Args:
            name:
            size:

        Returns:

        """
        data = np.random.randint(0, 100, size=size)
        df = pd.DataFrame(data, columns=list(string.ascii_uppercase[0:size[1]]))
        df.to_csv(path, index=None, sep=sep)
        return df

    def create_local_image_file(self, path):
        from imageio import imsave
        import numpy as np
        img = np.zeros([100, 100, 3], dtype=np.uint8)
        img.fill(255)
        imsave(path, img)
        return img

    def create_job(self, name):
        cells = []
        code = "print('hello')"
        cells.append(v4.new_code_cell(source=code))
        notebook = v4.new_notebook(cells=cells)
        # put the notebook
        return self.om.jobs.put(notebook, name)

    def get_package_path(self):
        basepath = os.path.join(os.path.dirname(sys.modules['omegaml'].__file__), 'example')
        pkgpath = os.path.abspath(os.path.join(basepath, 'demo', 'helloworld'))
        return pkgpath




from unittest import TestCase

from omegaml import Omega
from omegaml.tests.cli.scenarios import CliTestScenarios
from omegaml.tests.util import OmegaTestMixin


class CliModelsTest(CliTestScenarios, OmegaTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.om = Omega()
        self.clean()

    def test_cli_models_list(self):
        # start with an empty list
        self.cli('models list')
        expected = self.pretend_log([])
        self.assertLogSize('info', 1)
        self.assertLogEqual('info', expected)
        # create a model
        self.make_model('reg')
        self.cli('models list')
        expected = self.pretend_log(['reg'], start_empty=True)
        self.assertLogSize('info', 1)
        self.assertLogEqual('info', expected)
        # check we can get back a Metadata object
        self.cli('models list --raw')
        expected = self.pretend_log('Metadata(')
        self.assertLogContains('info', expected)

    def test_cli_models_put(self):
        self.cli(f'models put omegaml.example.demo.modelfn.create_model testmodel')
        expected = self.pretend_log('Metadata(name=testmodel')
        self.assertLogContains('info', expected)

    def test_cli_models_metadata(self):
        self.make_model('reg')
        self.cli('models metadata reg')
        expected = self.pretend_log('"kind": "sklearn.joblib"')
        self.assertLogContains('info', expected)
        expected = self.pretend_log('"name": "reg"')
        self.assertLogContains('info', expected)



